from __future__ import annotations

import pytest
import polars as pl
import polars_ds as pds
from polars.testing import assert_frame_equal, assert_series_equal


def test_replace_non_ascii():
    df = pl.DataFrame({"x": ["mercy", "xbĤ", "ĤŇƏ"]})

    assert_frame_equal(
        df.select(pds.replace_non_ascii("x")), pl.DataFrame({"x": ["mercy", "xb", ""]})
    )

    assert_frame_equal(
        df.select(pds.replace_non_ascii("x", "?")),
        pl.DataFrame({"x": ["mercy", "xb?", "???"]}),
    )

    assert_frame_equal(
        df.select(pds.replace_non_ascii("x", "??")),
        pl.DataFrame({"x": ["mercy", "xb??", "??????"]}),
    )


def test_remove_diacritics():
    df = pl.DataFrame({"x": ["mercy", "mèrcy", "françoise", "über"]})

    assert_frame_equal(
        df.select(pds.remove_diacritics("x")),
        pl.DataFrame({"x": ["mercy", "mercy", "francoise", "uber"]}),
    )


def test_normalize_string():
    df = pl.DataFrame({"x": ["\u0043\u0327"], "y": ["\u00c7"]}).with_columns(
        pl.col("x").eq(pl.col("y")).alias("is_equal"),
        pds.normalize_string("x", "NFC")
        .eq(pds.normalize_string("y", "NFC"))
        .alias("normalized_is_equal"),
    )

    assert df["is_equal"].sum() == 0
    assert df["normalized_is_equal"].sum() == df.height


def test_map_words():
    df = pl.DataFrame({"x": ["one two three", "onetwo three"]})

    assert_frame_equal(
        df.select(pds.map_words("x", {"two": "2"})),
        pl.DataFrame({"x": ["one 2 three", "onetwo three"]}),
    )

    assert_frame_equal(
        df.select(pds.map_words("x", {"two": "2", "three": "3"})),
        pl.DataFrame({"x": ["one 2 3", "onetwo 3"]}),
    )

    assert_frame_equal(
        df.select(pds.map_words("x", {"four": "4"})),
        pl.DataFrame({"x": ["one two three", "onetwo three"]}),
    )


def test_normalize_whitespace():
    df = pl.DataFrame({"x": ["a   b", "ab", "a b", "a\t\nb"]})

    assert_frame_equal(
        df.select(pds.normalize_whitespace("x")),
        pl.DataFrame({"x": ["a b", "ab", "a b", "a b"]}),
    )

    assert_frame_equal(
        df.select(pds.normalize_whitespace("x", only_spaces=True)),
        pl.DataFrame({"x": ["a b", "ab", "a b", "a\t\nb"]}),
    )


@pytest.mark.parametrize(
    "df, res",
    [
        (
            pl.DataFrame(
                {
                    "a": ["karolin", "karolin", "kathrin", "0000", "2173896"],
                    "b": ["kathrin", "kerstin", "kerstin", "1111", "2233796"],
                }
            ),
            pl.DataFrame({"a": pl.Series([3, 3, 4, 4, 3], dtype=pl.UInt32)}),
        ),
    ],
)
def test_hamming(df, res):
    assert_frame_equal(df.select(pds.str_hamming("a", pl.col("b"))), res)
    assert_frame_equal(df.select(pds.str_hamming("a", pl.col("b"), parallel=True)), res)
    assert_frame_equal(df.lazy().select(pds.str_hamming("a", pl.col("b"))).collect(), res)


@pytest.mark.parametrize(
    "df, res",
    [
        (  # From Wikipedia
            pl.DataFrame(
                {
                    "a": ["FAREMVIEL"],
                    "b": ["FARMVILLE"],
                }
            ),
            pl.DataFrame({"a": pl.Series([(1 / 3) * (16 / 9 + 7 / 8)], dtype=pl.Float64)}),
        ),
    ],
)
def test_jaro(df, res):
    assert_frame_equal(df.select(pds.str_jaro("a", pl.col("b"))), res)
    assert_frame_equal(df.select(pds.str_jaro("a", pl.col("b"), parallel=True)), res)
    assert_frame_equal(df.lazy().select(pds.str_jaro("a", pl.col("b"))).collect(), res)


@pytest.mark.parametrize(
    "df, res",
    [
        (  # From Wikipedia
            pl.DataFrame(
                {
                    "a": ["FAREMVIEL"],
                    "b": ["FARMVILLE"],
                }  # longest common sebsequence is 7. Distance is max(9, 9) - 7 = 2
            ),
            pl.DataFrame({"a": pl.Series([2], dtype=pl.UInt32)}),
        ),
    ],
)
def test_lcs_subseq_dist(df, res):
    assert_frame_equal(df.select(pds.str_lcs_subseq_dist("a", pl.col("b"), return_sim=False)), res)
    assert_frame_equal(
        df.select(pds.str_lcs_subseq_dist("a", pl.col("b"), return_sim=False, parallel=True)), res
    )
    assert_frame_equal(
        df.lazy().select(pds.str_lcs_subseq_dist("a", pl.col("b"), return_sim=False)).collect(), res
    )


@pytest.mark.parametrize(
    "a, b, lcs",
    [
        (
            ["ABCDEF", "abc", "common", "", "abcdefg", "你好世界", "🚀🛰️🌌"],
            ["ZBCDG", "xyz", "common", "test", "xabcdey", "世界和平", "🛰️✨🌟"],
            ["BCD", "", "common", "", "abcde", "世界", "🛰️"],
        ),
    ],
)
def test_lcs_substr(a, b, lcs):
    df = pl.DataFrame({"a": a, "b": b, "lcs": lcs}).with_columns(
        pds_lcs1=pds.str_lcs_substr("a", "b"), pds_lcs2=pds.str_lcs_substr("a", "b", parallel=True)
    )

    assert_series_equal(df["lcs"], df["pds_lcs1"], check_names=False)
    assert_series_equal(df["lcs"], df["pds_lcs2"], check_names=False)


@pytest.mark.parametrize(
    "a, b, lcs",
    [
        (
            ["AGGTAB", "abc", "你好世界和平", "🚀🛰️🌌"],
            ["GXTXAYB", "xyz", "世界a和平", "🛰️✨🌟"],
            ["GTAB", "", "世界和平", "🛰️"],
        ),
    ],
)
def test_lcs_subseq(a, b, lcs):
    df = pl.DataFrame({"a": a, "b": b, "lcs": lcs}).with_columns(
        pds_lcs1=pds.str_lcs_subseq("a", "b"), pds_lcs2=pds.str_lcs_subseq("a", "b", parallel=True)
    )

    assert_series_equal(df["lcs"], df["pds_lcs1"], check_names=False)
    assert_series_equal(df["lcs"], df["pds_lcs2"], check_names=False)


@pytest.mark.parametrize(
    "df, res",
    [
        (
            pl.DataFrame({"a": ["kitten", "mary", "may"], "b": ["sitting", "merry", "mayer"]}),
            pl.DataFrame({"a": pl.Series([3, 2, 2], dtype=pl.UInt32)}),
        ),
        (
            pl.DataFrame(
                {
                    "a": [
                        "Ostroróg",
                        "Hätönen",
                        "Kõivsaar",
                        "Pöitel",
                        "Vystrčil",
                        "Särki",
                        "Chreptavičienė",
                        "Väänänen",
                        "Führus",
                        "Könönen",
                        "Väänänen",
                        "Łaszczyński",
                        "Pärnselg",
                        "Könönen",
                        "Piątkowski",
                        "D’Amore",
                        "Körber",
                        "Särki",
                        "Kärson",
                        "Węgrzyn",
                    ],
                    "b": [
                        "Könönen",
                        "Hätönen",
                        "Wyżewski",
                        "Jäger",
                        "Hätönen",
                        "Mäns",
                        "Chreptavičienė",
                        "Väänänen",
                        "Ahısha",
                        "Jürist",
                        "Vainjärv",
                        "Łaszczyński",
                        "Pärnselg",
                        "Führus",
                        "Kübarsepp",
                        "Németi",
                        "Räheso",
                        "Käri",
                        "Jäger",
                        "Setälä",
                    ],
                }
            ),
            pl.DataFrame(
                {
                    "a": pl.Series(
                        [8, 0, 8, 5, 7, 4, 0, 0, 6, 7, 6, 0, 0, 7, 10, 6, 6, 2, 5, 7],
                        dtype=pl.UInt32,
                    )
                }
            ),
        ),
    ],
)
def test_levenshtein(df, res):
    assert_frame_equal(df.select(pds.str_leven("a", pl.col("b"))), res)

    assert_frame_equal(df.lazy().select(pds.str_leven("a", pl.col("b"))).collect(), res)


@pytest.mark.parametrize(
    "df, bound, res",
    [
        (
            pl.DataFrame(
                {"a": ["kitten", "mary", "may", None], "b": ["sitting", "merry", "mayer", ""]}
            ),
            2,
            pl.DataFrame({"a": pl.Series([False, True, True, None])}),
        ),
    ],
)
def test_levenshtein_filter(df, bound, res):
    assert_frame_equal(
        df.select(pds.filter_by_levenshtein(pl.col("a"), pl.col("b"), bound=bound)), res
    )

    assert_frame_equal(
        df.select(pds.filter_by_levenshtein(pl.col("a"), pl.col("b"), bound=bound, parallel=True)),
        res,
    )

    assert_frame_equal(
        df.lazy()
        .select(pds.filter_by_levenshtein(pl.col("a"), pl.col("b"), bound=bound))
        .collect(),
        res,
    )


@pytest.mark.parametrize(
    "df, bound, res",
    [
        (
            pl.DataFrame(
                {
                    "a": ["AAAAA", "AAATT", "AATTT", "AAAAA", "AAAAA"],
                    "b": ["AAAAT", "AAAAA", "ATATA", "AAAAA", "TTTTT"],
                }
            ),
            2,
            pl.DataFrame({"a": pl.Series([True, True, False, True, False])}),
        ),
    ],
)
def test_hamming_filter(df, bound, res):
    assert_frame_equal(df.select(pds.filter_by_hamming("a", pl.col("b"), bound=bound)), res)

    assert_frame_equal(
        df.select(pds.filter_by_hamming("a", pl.col("b"), bound=bound, parallel=True)),
        res,
    )

    assert_frame_equal(
        df.lazy().select(pds.filter_by_hamming("a", pl.col("b"), bound=bound)).collect(),
        res,
    )


@pytest.mark.parametrize(
    "a, word, res",
    [
        (["abc", "bbb", "ccc"], "abd", "abc"),
    ],
)
def test_nearest_str(a, word, res):
    df = pl.DataFrame(
        {
            "a": a,
        }
    )

    nearest = df.select(pds.str_nearest("a", word=word)).item(0, 0)
    assert res == nearest


@pytest.mark.parametrize(
    "df, res",
    [
        (
            pl.DataFrame({"a": ["CA", "AB", None], "b": ["ABC", "BA", "a"]}),
            pl.DataFrame({"a": pl.Series([3, 1, None], dtype=pl.UInt32)}),
        ),
    ],
)
def test_osa(df, res):
    assert_frame_equal(df.select(pds.str_osa("a", pl.col("b"))), res)

    assert_frame_equal(df.select(pds.str_osa("a", pl.col("b"), parallel=True)), res)
    assert_frame_equal(df.lazy().select(pds.str_osa("a", pl.col("b"))).collect(), res)


@pytest.mark.parametrize(
    "df, res",
    [
        (
            pl.DataFrame({"a": ["kitten"], "b": ["sitting"]}),
            pl.DataFrame({"a": pl.Series([4 / 11], dtype=pl.Float64)}),
        ),
    ],
)
def test_sorensen_dice(df, res):
    assert_frame_equal(df.select(pds.str_sorensen_dice("a", pl.col("b"))), res)

    assert_frame_equal(df.select(pds.str_sorensen_dice("a", pl.col("b"), parallel=True)), res)

    assert_frame_equal(df.lazy().select(pds.str_sorensen_dice("a", pl.col("b"))).collect(), res)


@pytest.mark.parametrize(
    "df, size, res",
    [
        (
            pl.DataFrame({"a": ["apple", "test", "moon"], "b": ["let", "tests", "sun"]}),
            2,
            pl.DataFrame({"a": pl.Series([0.2, 0.75, 0.0], dtype=pl.Float64)}),
        ),
        (
            pl.DataFrame({"a": ["apple", "test", "moon"], "b": ["let", "tests", "sun"]}),
            3,
            pl.DataFrame({"a": pl.Series([0.0, 2 / 3, 0.0], dtype=pl.Float64)}),
        ),
    ],
)
def test_str_jaccard(df, size, res):
    assert_frame_equal(df.select(pds.str_jaccard("a", pl.col("b"), substr_size=size)), res)
    assert_frame_equal(
        df.select(pds.str_jaccard("a", pl.col("b"), substr_size=size, parallel=True)),
        res,
    )
    assert_frame_equal(
        df.lazy()
        .select(pds.str_jaccard("a", pl.col("b"), substr_size=size, parallel=True))
        .collect(),
        res,
    )


@pytest.mark.parametrize(
    "df, size, alpha, beta, res",
    [
        (
            pl.DataFrame({"a": ["apple", "test", "moon"], "b": ["let", "tests", "sun"]}),
            2,
            0.5,
            0.5,
            pl.DataFrame(
                {"a": pl.Series([0.3333333333333333, 0.8571428571428571, 0.0], dtype=pl.Float64)}
            ),
        ),
        (
            pl.DataFrame({"a": ["apple", "test", "moon"], "b": ["let", "tests", "sun"]}),
            3,
            0.1,
            0.9,
            pl.DataFrame({"a": pl.Series([0.0, 0.6896551724137931, 0.0], dtype=pl.Float64)}),
        ),
    ],
)
def test_tversky(df, size, alpha, beta, res):
    assert_frame_equal(
        df.select(pds.str_tversky_sim("a", pl.col("b"), alpha=alpha, beta=beta, substr_size=size)),
        res,
    )
    assert_frame_equal(
        df.select(
            pds.str_tversky_sim(
                "a", pl.col("b"), alpha=alpha, beta=beta, substr_size=size, parallel=True
            )
        ),
        res,
    )
    assert_frame_equal(
        df.lazy()
        .select(
            pds.str_tversky_sim(
                "a", pl.col("b"), alpha=alpha, beta=beta, substr_size=size, parallel=True
            )
        )
        .collect(),
        res,
    )



"""Machine Learning / Time series Loss and Evaluation Metrics"""

from __future__ import annotations

import polars as pl
import math

# Internal dependencies
from polars_ds._utils import pl_plugin, to_expr
from polars_ds.typing import MultiAUCStrategy
from .num import add_at

__all__ = [
    "query_r2",
    "query_adj_r2",
    "query_log_cosh",
    "query_hubor_loss",
    "query_l1",
    "query_l2",
    "query_l_inf",
    "query_log_loss",
    "query_mape",
    "query_smape",
    "query_mase",
    "query_msle",
    "query_mad",
    "query_roc_auc",
    "query_tpr_fpr",
    "query_binary_metrics",
    "query_multi_roc_auc",
    "query_cat_cross_entropy",
    "query_confusion_matrix",
    "query_fairness",
    "query_p_pct_score",
    "query_mcc",
    "query_dcg_score",
    "query_ndcg_score",
]

# Confusion matrix based metrics should all be covered. If there is no
# specific function for it, we can find it in query_confusion_matrix.


def query_mad(x: str | pl.Expr, use_mean: bool = True) -> pl.Expr:
    """
    Computes the Mean/median Absolute Deviation.

    Parameters
    ----------
    x
        An expression represeting the actual
    use_mean
        If true, computes mean absolute deviation. If false, use median instead of mean.
    """
    xx = to_expr(x)
    if use_mean:
        return (xx - xx.mean()).abs().mean()
    else:
        return (xx - xx.median()).abs().median()


def query_r2(actual: str | pl.Expr, pred: str | pl.Expr) -> pl.Expr:
    """
    Returns the coefficient of determineation for a regression model.

    Parameters
    ----------
    actual
        An expression represeting the actual
    pred
        A Polars expression representing predictions
    """
    a = to_expr(actual)
    p = to_expr(pred)
    diff = a - p
    ss_res = diff.dot(diff)
    diff2 = a - a.mean()
    ss_tot = diff2.dot(diff2)
    return 1.0 - ss_res / ss_tot


def query_adj_r2(actual: str | pl.Expr, pred: str | pl.Expr, p: int) -> pl.Expr:
    """
    Returns the adjusted coefficient of determineation for a regression model.

    Parameters
    ----------
    actual
        An expression represeting the actual
    pred
        A Polars expression representing predictions
    p
        The number of explanatory variables
    """
    actual_expr = to_expr(actual)
    pred_expr = to_expr(pred)
    diff = actual_expr - pred_expr
    ss_res = diff.dot(diff)
    diff2 = actual_expr - actual_expr.mean()
    ss_tot = diff2.dot(diff2)
    df_res = actual_expr.len() - p
    df_tot = actual_expr.len() - 1
    return 1.0 - (ss_res / df_res) / (ss_tot / df_tot)


def query_log_cosh(actual: str | pl.Expr, pred: str | pl.Expr, normalize: bool = True) -> pl.Expr:
    """
    Computes log cosh of the the prediction error, which is a smooth variation of MAE (L1 loss).
    """
    a, p = to_expr(actual), to_expr(pred)
    if normalize:
        return (p - a).cosh().log().sum() / a.count()
    return (p - a).cosh().log().sum()


def query_hubor_loss(actual: str | pl.Expr, pred: str | pl.Expr, delta: float) -> pl.Expr:
    """
    Computes huber loss between this and the other expression. This assumes
    this expression is actual, and the input is predicted, although the order
    does not matter in this case.

    Parameters
    ----------
    pred
        An expression represeting the column with predicted probability.
    """
    a, p = to_expr(actual), to_expr(pred)
    temp = (a - p).abs()
    return (
        pl.when(temp <= delta).then(0.5 * temp.pow(2)).otherwise(delta * (temp - 0.5 * delta)).sum()
        / a.count()
    )


def query_l2(actual: str | pl.Expr, pred: str | pl.Expr, normalize: bool = True) -> pl.Expr:
    """
    Returns squared L2 loss, aka. mean squared error.

    Parameters
    ----------
    actual
        An expression represeting the actual
    pred
        A Polars expression representing predictions
    normalize
        Whether to divide by N.
    """
    a = to_expr(actual)
    p = to_expr(pred)
    diff = a - p
    if normalize:
        return diff.dot(diff) / a.count()
    return diff.dot(diff)


def query_l1(actual: str | pl.Expr, pred: str | pl.Expr, normalize: bool = True) -> pl.Expr:
    """
    Returns L1 loss, aka. mean absolute error.

    Parameters
    ----------
    actual
        An expression represeting the actual
    pred
        A Polars expression representing predictions
    normalize
        Whether to divide by N. Nulls won't be counted in N.
    """
    a = to_expr(actual)
    p = to_expr(pred)
    if normalize:
        return (a - p).abs().sum() / a.count()
    return (a - p).abs().sum()


def query_l_inf(actual: str | pl.Expr, pred: str | pl.Expr) -> pl.Expr:
    """
    Returns L Inf loss.

    Parameters
    ----------
    actual
        An expression represeting the actual
    pred
        A Polars expression representing predictions
    """
    a = to_expr(actual)
    p = to_expr(pred)
    return (a - p).abs().max()


def query_log_loss(actual: str | pl.Expr, pred: str | pl.Expr, normalize: bool = True) -> pl.Expr:
    """
    Computes log loss, aka binary cross entropy loss, between self and other `pred` expression.

    Parameters
    ----------
    actual
        An expression represeting the actual
    pred
        An expression represeting the column with predicted probability.
    normalize
        Whether to divide by N.
    """
    a = to_expr(actual).cast(pl.Float64)
    p = to_expr(pred).cast(pl.Float64)
    first = pl_plugin(
        args=[a, p],
        symbol="pl_xlogy",
        is_elementwise=True,
    )
    second = pl_plugin(
        args=[pl.lit(1.0, dtype=pl.Float64) - a, pl.lit(1.0, dtype=pl.Float64) - p],
        symbol="pl_xlogy",
        is_elementwise=True,
    )

    if normalize:
        return -(first + second).mean()
    return -(first + second).sum()


def query_mape(actual: str | pl.Expr, pred: str | pl.Expr, weighted: bool = False) -> pl.Expr:
    """
    Computes mean absolute percentage error between self and the other `pred` expression.
    If weighted, it will compute the weighted version as defined here:

    https://en.wikipedia.org/wiki/Mean_absolute_percentage_error

    Parameters
    ----------
    actual
        An expression represeting the actual
    pred
        An expression represeting the column with predicted probability.
    weighted
        If true, computes wMAPE in the wikipedia article
    """
    a = to_expr(actual)
    p = to_expr(pred)
    if weighted:
        return (a - p).abs().sum() / a.abs().sum()
    else:
        return (1 - p / a).abs().mean()


def query_smape(actual: str | pl.Expr, pred: str | pl.Expr) -> pl.Expr:
    """
    Computes symmetric mean absolute percentage error between self and other `pred` expression.
    The value is always between 0 and 1. This is the third version in the wikipedia without
    the 100 factor.

    https://en.wikipedia.org/wiki/Symmetric_mean_absolute_percentage_error

    Parameters
    ----------
    actual
        An expression represeting the actual
    pred
        A Polars expression representing predictions
    """
    a = to_expr(actual)
    p = to_expr(pred)
    numerator = (a - p).abs()
    denominator = a.abs() + p.abs()
    return (numerator / denominator).sum() / a.count()


def query_mase(
    actual: str | pl.Expr,
    pred: str | pl.Expr,
    train: str | pl.Expr | float,
    freq: int = 1,
    use_mean: bool = True,
) -> pl.Expr:
    """
    Computes the Mean/Median Absolute Scaled Error. This is the time series version in the reference article.

    Note: typically, train = pl.col('y').filter(pl.col('time') < T), and
    pred = pl.col('y_pred').filter(pl.col('time') >= T) and actual = pl.col('y').filter(pl.col('time') >= T)

    Parameters
    ----------
    actual
        An expression represeting the actual
    pred
        A Polars expression representing predictions
    train
        A polars exression representing training data values. If train is a float, it is treated
        as the precomputed naive one-step forecast loss on training as in the definition.
    freq
        Defaults to 1 which applies to non-seasonal data, and you may set it to m (>0)
        which indicates the length of the season. How frequent does the period repeat itself? Every `freq`
        records.
    use_mean
        If true, this will compute Mean Absolute Scaled Error. If false, this uses median instead of mean.

    Reference
    ---------
    https://en.wikipedia.org/wiki/Mean_absolute_scaled_error
    """
    if freq < 1:
        raise ValueError("Input `freq` must be >= 1.")

    a: pl.Expr = to_expr(actual)
    p: pl.Expr = to_expr(pred)

    if isinstance(train, float):
        if use_mean:
            numerator = (a - p).abs().mean()
        else:
            numerator = (a - p).abs().median()

        return numerator / pl.lit(train)

    else:
        train_expr = to_expr(train)
        if use_mean:
            numerator = (a - p).abs().mean()
            denom = train_expr.diff(n=freq).abs().mean()
        else:
            numerator = (a - p).abs().median()
            denom = train_expr.diff(n=freq).abs().median()

        return numerator / denom


def query_msle(actual: str | pl.Expr, pred: str | pl.Expr, normalize: bool = True) -> pl.Expr:
    """
    Computes the mean square log error between this and the other `pred` expression.

    Parameters
    ----------
    pred
        An expression represeting the column with predicted probability.
    normalize
        If true, divide the result by length of the series
    """
    a = to_expr(actual)
    p = to_expr(pred)
    diff = a.log1p() - p.log1p()
    out = diff.dot(diff)
    if normalize:
        return out / a.count()
    return out


def query_roc_auc(
    actual: str | pl.Expr,
    pred: str | pl.Expr,
) -> pl.Expr:
    """
    Computes ROC AUC using self as actual and pred as predictions.

    Self must be binary and castable to type UInt32. If self is not all 0s and 1s or not binary,
    the result will not make sense, or some error may occur. If no positive class exist in data,
    NaN will be returned.

    Parameters
    ----------
    actual
        An expression represeting the actual. Must be castable to UInt32.
    pred
        An expression represeting the column with predicted probability.
    """
    return pl_plugin(
        symbol="pl_roc_auc",
        args=[to_expr(actual).cast(pl.UInt32), to_expr(pred).cast(pl.Float64)],
        returns_scalar=True,
    )


def query_tpr_fpr(
    actual: str | pl.Expr,
    pred: str | pl.Expr,
) -> pl.Expr:
    """
    Returns the TPR and FPR for all thresholds. This is useful when you want to study the thresholds
    or when you want to plot roc auc curve.

    Parameters
    ----------
    actual
        An expression represeting the actual. Must be castable to UInt32.
    pred
        An expression represeting the column with predicted probability.
    """
    return pl_plugin(
        symbol="pl_tpr_fpr",
        args=[to_expr(actual).cast(pl.UInt32), to_expr(pred).cast(pl.Float64)],
    )


def query_gini(actual: str | pl.Expr, pred: str | pl.Expr) -> pl.Expr:
    """
    Computes the Gini coefficient. This is 2 * AUC - 1.

    Self must be binary and castable to type UInt32. If self is not all 0s and 1s or not binary,
    the result will not make sense, or some error may occur. If no positive class exist in data,
    NaN will be returned.

    Parameters
    ----------
    actual
        An expression represeting the actual
    pred
        An expression represeting the column with predicted probability.
    """
    return query_roc_auc(actual, pred) * 2.0 - 1.0


def query_confusion_matrix(
    actual: str | pl.Expr,
    pred: str | pl.Expr,
    threshold: float = 0.5,
    all_metrics: bool = False,
) -> pl.Expr:
    """
    Computes the binary confusion matrix given the true labels (`actual`) and
    the predicted labels (computed from `pred`, a column of predicted scores and
    `threshold`). When a divide by zero is encountered, NaN is returned.

    Parameters
    ----------
    actual : str | pl.Expr
        An expression representing the actual labels. Must be castable to boolean
    pred : str | pl.Expr
        An expression representing the column with predicted probability
    threshold : float, optional
        The threshold used to compute the predicted labels, by default 0.5
    all_metrics : bool, optional
        If True, compute all 25 possible confusion matrix statistics instead of
        just True Positive, False Positive, True Negative, False Negative,
        by default False

    Returns
    -------
    pl.Expr
        A struct of confusion matrix metrics

    Examples
    --------
    Limited to just the basic confusion matrix

    >>> df = pl.DataFrame({"actual": [1, 0, 1], "pred": [0.4, 0.6, 0.9]})
    >>> df.select(pds.query_confusion_matrix("actual", "pred").alias("metrics")).unnest("metrics")
    shape: (1, 4)
    ┌─────┬─────┬─────┬─────┐
    │ tn  ┆ fp  ┆ fn  ┆ tp  │
    │ --- ┆ --- ┆ --- ┆ --- │
    │ u32 ┆ u32 ┆ u32 ┆ u32 │
    ╞═════╪═════╪═════╪═════╡
    │ 0   ┆ 1   ┆ 1   ┆ 1   │
    └─────┴─────┴─────┴─────┘

    With `all_metrics` set to True

    >>> df.select(
    ...     pds.query_confusion_matrix("actual", "pred", all_metrics=True).alias("metrics")
    ... ).unnest("metrics")
    shape: (1, 25)
    ┌─────┬─────┬─────┬─────┬───┬────────────┬─────┬─────┬─────┐
    │ tn  ┆ fp  ┆ fn  ┆ tp  ┆ … ┆ markedness ┆ fdr ┆ npv ┆ dor │
    │ --- ┆ --- ┆ --- ┆ --- ┆   ┆ ---        ┆ --- ┆ --- ┆ --- │
    │ u32 ┆ u32 ┆ u32 ┆ u32 ┆   ┆ f64        ┆ f64 ┆ f64 ┆ f64 │
    ╞═════╪═════╪═════╪═════╪═══╪════════════╪═════╪═════╪═════╡
    │ 0   ┆ 1   ┆ 1   ┆ 1   ┆ … ┆ -0.5       ┆ 0.5 ┆ 0.0 ┆ NaN │
    └─────┴─────┴─────┴─────┴───┴────────────┴─────┴─────┴─────┘
    """
    # Cast to bool first to check the label is in correct format. Then back to u32.
    act = to_expr(actual).cast(pl.Boolean).cast(pl.UInt32)
    p = to_expr(pred).gt(threshold).cast(pl.UInt32)
    res = pl_plugin(
        symbol="pl_binary_confusion_matrix",
        args=[(2 * act) + p],  # See Rust code for bincount trick
        returns_scalar=True,
    )
    if all_metrics:
        return res
    else:
        return pl.struct(
            res.struct.field("tn"),
            res.struct.field("fp"),
            res.struct.field("fn"),
            res.struct.field("tp"),
        )


def query_binary_metrics(
    actual: str | pl.Expr, pred: str | pl.Expr, threshold: float = 0.5
) -> pl.Expr:
    """
    Computes the following binary classificaition metrics using self as actual and pred as predictions:
    precision, recall, f, average_precision and roc_auc. The return will be a struct with values
    having the names as given here.

    Self must be binary and castable to type UInt32. If self is not all 0s and 1s,
    the result will not make sense, or some error may occur. If there is no positive class in data,
    NaN or other numerical error may occur.

    Average precision is computed using Sum (R_n - R_n-1)*P_n-1, which is not the textbook definition,
    but is consistent with Scikit-learn. For more information, see
    https://scikit-learn.org/stable/modules/generated/sklearn.metrics.average_precision_score.html

    Parameters
    ----------
    actual
        An expression represeting the actual
    pred
        An expression represeting the column with predicted probability.
    threshold
        The threshold used to compute precision, recall and f (f score).
    """
    return pl_plugin(
        symbol="pl_combo_b",
        args=[
            to_expr(actual).cast(pl.UInt32),
            to_expr(pred),
            pl.lit(threshold, dtype=pl.Float64),
        ],
        returns_scalar=True,
    )


def query_multi_roc_auc(
    actual: str | pl.Expr,
    pred: str | pl.Expr,
    n_classes: int,
    strategy: MultiAUCStrategy = "weighted",
) -> pl.Expr:
    """
    Computes multiclass ROC AUC. Self (actuals) must be labels represented by integer values
    ranging in the range [0, n_classes), and pred must be a column of list[f64] with size `n_classes`.

    Parameters
    ----------
    actual
        An expression represeting the actual
    pred
        The multilabel prediction column
    n_classes
        The number of classes
    strategy
        Either `macro` or `weighted`, which are defined the same as in Scikit-learn.
    """
    a = to_expr(actual)
    p = to_expr(pred)
    if strategy == "macro":
        actuals = [a == i for i in range(n_classes)]
        preds = [p.list.get(i) for i in range(n_classes)]
        return pl.sum_horizontal(query_roc_auc(a, p) for a, p in zip(actuals, preds)) / n_classes
    elif strategy == "weighted":
        actuals = [a == i for i in range(n_classes)]
        preds = [p.list.get(i) for i in range(n_classes)]
        return (
            pl.sum_horizontal(a.sum() * query_roc_auc(a, p) for a, p in zip(actuals, preds))
            / pl.len()
        )
    else:
        raise NotImplementedError


def query_cat_cross_entropy(
    actual: str | pl.Expr, pred: str | pl.Expr, normalize: bool = True, dense: bool = True
) -> pl.Expr:
    """
    Returns the categorical cross entropy. If you want to avoid numerical error due to log, please
    set pred = pred + epsilon.

    Parameters
    ----------
    actual
        An expression represeting the actual
    pred
        An expression represeting the predicted probabilities for the classes. Must of be List/arr[f64] type.
    normalize
        Whether to divide by N.
    dense
        If true, actual has to be a dense vector (a single number for each row, starting from 0). If false, it has
        to be a column of lists/arrs with only one 1 and 0s otherwise.
    """
    a = to_expr(actual)
    p = to_expr(pred)
    if dense:
        y_prob = p.list.get(a)
    else:
        y_prob = p.list.get(a.list.arg_max())
    if normalize:
        return -y_prob.log().sum() / a.count()
    return -y_prob.log().sum()


def query_mcc(y_true: str | pl.Expr, y_pred: str | pl.Expr) -> pl.Expr:
    """
    Returns the Matthews correlation coefficient (phi coefficient). The inputs must be 0s and 1s
    and castable to u32. If not, the result may not be correct. See query_confusion_matrix for querying
    all the confusion metrics at the same time.

    Parameters
    ----------
    y_true
        The true labels. Must be 0s and 1s.
    y_pred
        The predicted labels. Must be 0s and 1s. E.g. This could be say (y_prob > 0.5).cast(pl.UInt32)

    Reference
    ---------
    https://en.wikipedia.org/wiki/Phi_coefficient
    """

    y = to_expr(y_true)
    x = to_expr(y_pred)
    combined = (2 * y + x).cast(pl.UInt32)

    return pl_plugin(
        symbol="pl_mcc",
        args=[combined],
        returns_scalar=True,
    )


def query_fairness(pred: str | pl.Expr, sensitive_cond: pl.Expr) -> pl.Expr:
    """
    A simple fairness metric for regression output. Computes the absolute difference between
    the average of the `pred` values on when the `sensitive_cond` is true vs the
    avg of the values when `sensitive_cond` is false.

    The lower this value is, the more fair is the model on the sensitive condition.

    Parameters
    ----------
    pred
        The predictions
    sensitive_cond
        A boolean expression representing the sensitive condition
    """
    p = to_expr(pred)
    return (p.filter(sensitive_cond).mean() - p.filter(~sensitive_cond).mean()).abs()


def query_p_pct_score(pred: str | pl.Expr, sensitive_cond: pl.Expr) -> pl.Expr:
    """
    Computes the 'p-percent score', which measures the fairness of a classification
    model on a sensitive_cond. Let z = the sensitive_cond, then:

    p-percent score = min(P(y = 1 | z = 1) / P(y = 1 | z = 0), P(y = 1 | z = 0) / P(y = 1 | z = 1))

    Parameters
    ----------
    pred
        The predictions. Must be 0s and 1s.
    sensitive_cond
        A boolean expression representing the sensitive condition
    """
    p = to_expr(pred)
    p_y1_z1 = p.filter(
        sensitive_cond
    ).mean()  # since p is 0s and 1s, this is equal to P(pred = 1 | sensitive_cond)
    p_y1_z0 = p.filter(~sensitive_cond).mean()
    ratio = p_y1_z1 / p_y1_z0
    return pl.min_horizontal(ratio, 1 / ratio)


def query_dcg_score(
    y_true: str | pl.Expr,
    y_score: str | pl.Expr,
    log_base: float = 2.0,
    k: int | None = None,
    ignore_ties: bool = False,
) -> pl.Expr:
    """
    Parameters
    ----------
    y_true:
        This is often called the `relevance` column. The relevance score represents the
        true relevance/importance of each item.
    score_col:
        The name/expr of the column containing the predicted scores used for ranking the items.
    log_base:
        The log base used in the discount factor
    k:
        The number of top items to consider in the NDCG calculation. If None, all items are
        considered. Defaults to None.
    ignore_ties:
        If True, handles tied scores by averaging their contributions. If False, ranks items with
        the same score sequentially. Defaults to False.
    """

    yt = to_expr(y_true)
    ys = to_expr(y_score)

    range_ = pl.int_range(1, pl.len() + 1)
    discount = math.log(log_base) / range_.log1p()
    if k is not None:
        discount = discount * (range_ <= k).cast(pl.Float64)

    if ignore_ties:
        ranking = ys.arg_sort().reverse()
        return (yt.gather(ranking)).dot(discount)
    else:
        discount_cumsum = discount.cum_sum()
        # inv, and counts are equivalent to
        # _, inv, counts = np.unique(-y_score, return_inverse=True, return_counts=True)
        counts = ys.sort(descending=True).unique_counts()
        inv = (ys.rank(method="dense", descending=True) - 1).set_sorted(descending=True)
        # The add_at function is an equivalent to numpy.add.at
        n_unique = counts.len()
        ranked = add_at(inv, yt, n_unique) / counts
        groups = counts.cum_sum() - 1

        # PL VERSION
        major, minor, patch = list(map(lambda v: int(v), pl.__version__.split(".")))
        if (major, minor, patch) < (1, 32, 3):
            discount_sums = (
                discount_cumsum.gather(groups)
                .diff()
                .fill_null(discount_cumsum.get(groups.first()).first())
            )  # The only possible null is at position 0

        else:  # in newer versions, get() returns a value which auto resolves to a scaler
            discount_sums = (
                discount_cumsum.gather(groups).diff().fill_null(discount_cumsum.get(groups.first()))
            )  # The only possible null is at position 0

        return ranked.dot(discount_sums)


def query_ndcg_score(
    y_true: str | pl.Expr,
    y_score: str | pl.Expr,
    k: int | None = None,
    ignore_ties: bool = False,
) -> pl.Expr:
    """Compute Normalized Discounted Cumulative Gain.

    NDCG is a measure of ranking quality that considers both the relevance of items
    and their positions in the ranked list. It normalizes the DCG score by the ideal
    DCG score (the maximum possible DCG for a given set of relevance scores).

    Note: (1) ndcg_score should not be used on negative y_true values. (2) This function
    might run faster in lazy Polars.

    Parameters
    ----------
    y_true:
        This is often called the `relevance` column. The relevance score represents the
        true relevance/importance of each item.
    score_col:
        The name/expr of the column containing the predicted scores used for ranking the items.
    k:
        The number of top items to consider in the NDCG calculation. If None, all items are
        considered. Defaults to None.
    ignore_ties:
        If True, handles tied scores by averaging their contributions. If False, ranks items with
        the same score sequentially. Defaults to False.

    Examples
    --------
    Basic usage with a simple ranking task:
    >>> df = pl.DataFrame(
    ...     {
    ...         "query_id": [1, 1, 1, 2, 2],
    ...         "relevance": [3, 2, 1, 2, 1],
    ...         "score": [0.9, 0.8, 0.7, 0.6, 0.5],
    ...     }
    ... )
    >>> ndcg = (
    ...     df.group_by("query_id")
    ...     .agg(ndsg_score=pds.query_ndcg_score("relevance", "score", k=2, ignore_ties=False))
    ...     .select(pl.col("ndsg_score").mean())
    ... )
    >>> print(ndcg)  # Shows the mean NDCG@2 across all queries

    Handling tied scores:
    >>> df = pl.DataFrame(
    ...     {
    ...         "query_id": [1, 1, 1, 1],
    ...         "relevance": [3, 2, 2, 1],
    ...         "score": [0.9, 0.8, 0.8, 0.7],  # Note the tied scores
    ...     }
    ... )
    >>> # (no need to group by because it is a single query)
    >>> ndcg_no_tie_handling = df.select(ndcg=pds.query_ndcg_score("relevance", "score", k=3))
    >>> ndcg_tie_handling = df.select(
    ...     ndcg=pds.query_ndcg_score("relevance", "score", k=3, ignore_ties=True)
    ... )

    Using with recommendation systems:
    >>> recommendations = pl.DataFrame(
    ...     {
    ...         "user_id": [1, 1, 1, 2, 2, 2],
    ...         "item_id": ["i1", "i2", "i3", "i4", "i5", "i6"],
    ...         "relevance": [5, 4, 3, 4, 3, 2],  # User ratings or engagement metrics
    ...         "pred_score": [0.95, 0.85, 0.75, 0.9, 0.8, 0.7],  # Model predictions
    ...     }
    ... )
    >>> ndcg = recommendations.group_by("user_id").agg(
    ...     ndcg=pds.query_ndcg_score("relevance", "pred_score", k=10)
    ... )
    """
    gain = query_dcg_score(y_true, y_score, k=k, ignore_ties=ignore_ties)
    # Lazy Polars should be able to figure out things like
    # `discount` and `yt.sort(descending=True)` are common in the calculation of gain
    # and normalized_gain, and so should be only computed once.
    yt = to_expr(y_true)
    range_ = pl.int_range(1, pl.len() + 1)
    discount = math.log(2.0) / range_.log1p()
    if k is not None:
        discount = discount * (range_ <= k).cast(pl.Float64)
    # this is an optimization we can do because we know yt = ys. No need to compute the
    # ranking variable in the ignore_ties case
    normalizing_gain = (yt.sort(descending=True)).dot(discount)
    # since gain and normalizing_gain are a scalar, NaN means normalizing gain is 0
    return (gain / normalizing_gain).fill_nan(pl.lit(0.0, dtype=pl.Float64))



"""Machine Learning / Time series Pipelines with native Polars support."""

from __future__ import annotations

import polars as pl
import json
import sys
import polars.selectors as cs
from polars._typing import IntoExprColumn
from functools import partial
from typing import List, Dict, Any, Literal
from dataclasses import dataclass
from polars_ds._utils import to_expr

if sys.version_info >= (3, 11):
    from typing import Self
else:  # 3.10, 3.9, 3.8
    from typing_extensions import Self

# Internal Depenedncies
from . import transforms as t
from ._step import (
    PLContext,
    ExprStep,
    SortStep,
    SQLStep,
    GroupByAggStep,
    GroupByDynAggStep,
    FitStep,
    PipelineStep,
    MakeStep,
    _SERIALIZABLE_STEPS,
)
from polars_ds.typing import (
    PolarsFrame,
    SimpleImputeMethod,
    SimpleScaleMethod,
    QuantileMethod,
    EncoderDefaultStrategy,
)

__all__ = ["Pipeline", "Blueprint", "PLContext", "ExprStep"]


@dataclass
class StepRepr:
    """
    A representation of a step
    """

    name: str
    args: List[Any]
    kwargs: Dict[str, Any]

    @staticmethod
    def from_dict(dictionary: Dict[str, Any]) -> "StepRepr":
        try:
            name: str = dictionary["name"]
            args: List[Any] = dictionary.get("args", [])
            kwargs: Dict[str, Any] = dictionary.get("kwargs", {})
            if not isinstance(name, str):
                raise ValueError("Value of `name` must be a string.")
            if not isinstance(args, list):
                raise ValueError("Value of `args` must be a list.")
            if not isinstance(kwargs, dict):
                raise ValueError("Value of `kwargs` must be a dict.")
            if not all(isinstance(s, str) for s in kwargs.keys()):
                raise ValueError("All keys in `kwargs` must be strings.")
            return StepRepr(name=name, args=args, kwargs=kwargs)
        except Exception as e:
            raise ValueError(f"Keys missing or data type is not expected. Original error: \n{e}")


@dataclass
class Pipeline:
    """
    A ML/data transform pipeline. Pipelines should always come from the materialize call from a
    blueprint. In other words, a pipeline is a fitted blueprint.
    """

    name: str
    feature_names_in_: List[str]
    feature_names_out_: List[str]
    transforms: List[PipelineStep]
    ensure_features_in: bool = False
    ensure_features_out: bool = True
    lowercase: bool = False
    uppercase: bool = False

    def __str__(self) -> str:
        return self.transforms.__str__()

    def _get_init_plan(self, df: PolarsFrame) -> pl.LazyFrame:
        """
        Get an initial plan without any pipeline transforms.
        """
        if self.lowercase:
            plan = df.lazy().select(pl.all().name.to_lowercase())
        else:
            if self.uppercase:
                plan = df.lazy().select(pl.all().name.to_uppercase())
            else:
                plan = df.lazy()

        return plan

    def _generate_lazy_plan(self, df: PolarsFrame) -> pl.LazyFrame:
        """
        Generates a lazy plan for the incoming df

        Paramters
        ---------
        df
            If none, create the plan for the df that the pipe is initialized with. Otherwise,
            create the plan for the incoming df.
        """
        plan = self._get_init_plan(df)
        for step in self.transforms:
            plan = step.apply_df(plan)
        return plan

    def with_features_out(self, features: List[str], ensure_features_out: bool = True) -> Self:
        self.feature_names_out_ = list(features)
        self.ensure_features_out = ensure_features_out

    def to_json(self, path: str | None = None, **kwargs) -> str | None:
        """
        Turns self into a JSON string.

        Parameters
        ----------
        path
            If none, will return a json string. If given, this will be used as the path
            to save the pipeline and None will be returned.
        kwargs
            Keyword arguments to Python's default json
        """
        # Maybe support other json package?
        d = {
            "name": str(self.name),
            "feature_names_in_": list(self.feature_names_in_),
            "feature_names_out_": list(self.feature_names_out_),
            "transforms": [step.to_json() for step in self.transforms],
            "ensure_features_in": self.ensure_features_in,
            "ensure_features_out": self.ensure_features_out,
            "lowercase": self.lowercase,
            "uppercase": self.uppercase,
        }
        if path is None:
            return json.dumps(d, **kwargs)
        else:
            with open(path, "w") as f:
                json.dump(d, f)

            return None

    @staticmethod
    def from_json(json_str: str | bytes) -> "Pipeline":
        """
        Recreates a pipeline from a dictionary created by the `to_json` call.
        """
        pipeline_dict: Dict = json.loads(json_str)
        try:
            name: str = pipeline_dict["name"]
            transforms: List[str] = pipeline_dict["transforms"]
            feature_names_in_: List[str] = pipeline_dict["feature_names_in_"]
            feature_names_out_: List[str] = pipeline_dict["feature_names_out_"]
            ensure_features_in: bool = pipeline_dict["ensure_features_in"]
            ensure_features_out: bool = pipeline_dict["ensure_features_out"]
            lowercase: bool = pipeline_dict.get("lowercase", False)
            uppercase: bool = pipeline_dict.get("uppercase", False)
        except Exception as e:
            raise ValueError(f"Input dictionary is missing keywords. Original error: \n{e}")

        return Pipeline(
            name=name,
            feature_names_in_=feature_names_in_,
            feature_names_out_=feature_names_out_,
            transforms=[MakeStep.make(json.loads(step_str)) for step_str in transforms],
            ensure_features_in=ensure_features_in,
            ensure_features_out=ensure_features_out,
            lowercase=lowercase,
            uppercase=uppercase,
        )

    def ensure_features_io(self, ensure_in: bool = True, ensure_out: bool = True) -> Self:
        """
        Whether or not this pipeline should check the features coming in and out during transform.

        Parameters
        ----------
        ensure_in
            If true, the input df (during transform) must have the exact same features
            as when this pipeline was fitted (blueprint.materialize()). Setting this false means the input may
            have additional columns, and so this adds flexibility in the pipeline.
        ensure_out
            If true, only the output features during blueprint.materialize() will be kept at the end of the
            pipeline.
        """
        self.ensure_features_in = ensure_in
        self.ensure_features_out = ensure_out
        return self

    def transform(
        self, df: PolarsFrame, return_lazy: bool = False, set_features_out: bool = False, **kwargs
    ) -> PolarsFrame:
        """
        Transforms the df using the learned expressions.

        Paramters
        ---------
        df
            If none, transform the df that the pipe is initialized with. Otherwise, perform
            the learned transformations on the incoming df.
        return_lazy
            If true, return the lazy plan for the transformations
        set_features_out
            If true, set `self.feature_names_out_` to the output features from this transform run.
        **kwargs
            When return_lazy is True, kwargs will be passed to polars.LazyFrame.collect()
        """
        if self.ensure_features_in:
            if self.lowercase:
                columns = [c.lower() for c in df.lazy().collect_schema().names()]
            elif self.uppercase:
                columns = [c.upper() for c in df.lazy().collect_schema().names()]
            else:
                columns = df.lazy().collect_schema().names()
            extras = [c for c in columns if c not in self.feature_names_in_]
            missing = [c for c in self.feature_names_in_ if c not in columns]
            if len(extras) > 0 or len(missing) > 0:
                raise ValueError(
                    f"Input df doesn't have the features expected. Extra columns: {extras}. Missing columns: {missing}"
                )

        plan = self._generate_lazy_plan(df)
        if self.ensure_features_out:
            plan = plan.select(self.feature_names_out_)

        if set_features_out:
            self.feature_names_out_ = plan.collect_schema().names()

        # Add config here if streaming is needed
        return plan if return_lazy else plan.collect(**kwargs)


class Blueprint:
    """
    Blueprints for a ML/data transformation pipeline. In other words, this is a description of
    what a pipeline will be. No learning/fitting is done until self.materialize() is called.

    If the input df is lazy, the pipeline will collect at the time of fit.

    Note: although polars selectors work for most transformations and in most cases, it is still
    recommended that the user should use explicit expressions instead of selectors for most transformations.
    """

    def __init__(
        self,
        df: PolarsFrame,
        name: str = "test",
        target: str | None = None,
        exclude: List[str] | None = None,
        lowercase: bool = False,
        uppercase: bool = False,
    ):
        """
        Creates a blueprint object.

        Parameters
        ----------
        df
            Either a lazy or an eager Polars dataframe
        name
            Name of the blueprint.
        target
            Optionally indicate the target column in the ML pipeline. This will automatically prevent any transformation
            from changing the target column. (To be implemented: this should also automatically fill any transformation
            that requires a target name)
        exclude
            Any other column to exclude from global transformation. Note: this is only needed if you are not specifiying
            the exact columns to transform. E.g. when you are using a selector like cs.numeric() for all numeric columns.
            If this is the case and target is not set nor excluded, then the transformation may be applied to the target
            as well, which is not desired in most cases. Therefore, it is highly recommended you initialize with target name.
        lowercase
            Whether to insert a lowercase column name step before all other transformations.
            This takes precedence over uppercase.
        uppercase
            Whether to insert a uppercase column name step before all other transformations.
            This only happens if lowercase is False
        """

        self._df: pl.LazyFrame = df.lazy()
        if lowercase:
            self._df = self._df.select(pl.all().name.to_lowercase())
        else:
            if uppercase:
                self._df = self._df.select(pl.all().name.to_uppercase())

        self.name: str = str(name)
        self.target = target
        self.feature_names_in_: list[str] = self._df.collect_schema().names()

        self._steps: List[ExprStep | FitStep] = []
        self.exclude: List[str] = [] if target is None else [target]
        if exclude is not None:  # dedup in case user accidentally puts the same column name twice
            self.exclude = list(set(self.exclude + exclude))

        self.lowercase = lowercase
        self.uppercase = uppercase

    def __str__(self) -> str:
        out: str = ""
        out += f"Blueprint name: {self.name}\n"
        if self.lowercase:
            out += "Column names: Lowercase all incoming columns.\n"
        elif self.uppercase:
            out += "Column names: Uppercase all incoming columns.\n"

        out += f"Blueprint current steps: {len(self._steps)}\n"
        out += f"Features Expected: {self.feature_names_in_}\n"
        return out

    def _get_target(self, target: str | pl.Expr | None = None) -> str | pl.Expr:
        if target is None:
            if self.target is None:
                raise ValueError(
                    "Target is not given and blueprint is not initialized with a target."
                )
            return self.target
        else:
            return target

    def filter(self, by: str | pl.Expr) -> Self:
        """
        Filters on the dataframe using native polars expressions or SQL boolean expressions.

        Parameters
        ----------
        by
            Native polars boolean expression or SQL strings
        """
        self._steps.append(
            ExprStep(by if isinstance(by, pl.Expr) else pl.sql_expr(by), PLContext.FILTER)
        )
        return self

    def sql_transform(self, sql: str) -> Self:
        """
        Runs the SQL on the dataframe when it reaches this step. The user must ensure that
        the SQL is valid Polars SQL and all columns referred in the SQL exist at this point.
        The name "df" should be used to refer to the current state of the dataframe in the SQL.
        E.g. select * from df where A is True.

        Parameters
        ----------
        sql
            The SQL to run on the dataframe. Note: this step doesn't immedinately check the validity of
            the SQL statement.
        """
        self._steps.append(SQLStep(sql_str=sql))
        return self

    def cast_bools(self, dtype: pl.DataType = pl.UInt8) -> Self:
        """
        Cast all boolean columns in the dataframe to the given type.
        """
        self._steps.append(ExprStep(cs.boolean().cast(dtype), PLContext.WITH_COLUMNS))
        return self

    def impute(self, cols: IntoExprColumn, method: SimpleImputeMethod = "mean") -> Self:
        """
        Imputes null values in the given columns. Note: this doesn't fill NaN. If filling for NaN is needed,
        please manually do that.

        Parameters
        ----------
        cols
            Any Polars expression that can be understood as columns.
        method
            One of `mean`, `median`, `mode`. If `mode`, a random value will be chosen if there is
            a tie.
        """
        self._steps.append(FitStep(partial(t.impute, method=method), cols, self.exclude))
        return self

    def conditional_impute(
        self, rules_dict: Dict[str, str | pl.Expr], method: SimpleImputeMethod = "mean"
    ) -> Self:
        """
        Conditionally imputes values in the given columns. This transform will collect if input is lazy.

        Parameters
        ----------
        rules_dict
            Dictionary where keys are column names (must be string), and values are SQL/Polars Conditions
            that when true, those values in the column will be imputed,
            and the value to impute will be learned on the data where the condition is false.
        method
            One of `mean`, `median`, `mode`. If `mode`, a random value will be chosen if there is
            a tie.
        """
        self._steps.append(
            FitStep(
                partial(t.conditional_impute, rules_dict=rules_dict, method=method),
                None,
                self.exclude,
            )
        )
        return self

    def nan_to_null(self) -> Self:
        """
        Maps NaN values in all columns to null.
        """
        self._steps.append(ExprStep(cs.float().fill_nan(None), PLContext.WITH_COLUMNS))
        return self

    def int_to_float(self, f32: bool = True) -> Self:
        """
        Maps all integer columns to float.

        Parameters
        ----------
        f32
            If true, map all integer columns to f32 columns. Otherwise they will be
            casted to f64 columns.
        """
        if f32:
            self._steps.append(ExprStep(cs.integer().cast(pl.Float32), PLContext.WITH_COLUMNS))
        else:
            self._steps.append(ExprStep(cs.integer().cast(pl.Float64), PLContext.WITH_COLUMNS))
        return self

    def linear_impute(
        self, features: IntoExprColumn, target: str | pl.Expr | None = None, add_bias: bool = False
    ) -> Self:
        """
        Imputes the target column by training a simple linear regression using the other features. This will
        cast the target column to f64.

        Note: The linear regression will skip nulls whenever there is a null in the features or in the target.
        Additionally, if NaN or Inf exists in data, the linear regression result may be invalid or an error
        will be thrown. It is recommended to use this only after imputing and dealing with NaN and Infs for
        all feature columns first.

        Parameters
        ----------
        features
            Any Polars expression that can be understood as numerical columns which will be used as features
        target
            The target column
        add_bias
            Whether to add a bias term to the linear regression
        """
        self._steps.append(
            FitStep(
                partial(t.linear_impute, target=self._get_target(target), add_bias=add_bias),
                features,
                self.exclude,
            )
        )
        return self

    def scale(self, cols: IntoExprColumn, method: SimpleScaleMethod = "standard") -> Self:
        """
        Scales values in the given columns.

        Parameters
        ----------
        cols
            Any Polars expression that can be understood as columns.
        method
            One of `standard`, `min_max`, `abs_max`
        """
        self._steps.append(FitStep(partial(t.scale, method=method), cols, self.exclude))
        return self

    def robust_scale(self, cols: IntoExprColumn, q_low: float, q_high: float) -> Self:
        """
        Performs robust scaling on the given columns

        Paramters
        ---------
        cols
            Any Polars expression that can be understood as columns.
        q_low
            The lower quantile value
        q_high
            The higher quantile value
        """
        self._steps.append(
            FitStep(partial(t.robust_scale, q_low=q_low, q_high=q_high), cols, self.exclude)
        )
        return self

    def center(self, cols: IntoExprColumn) -> Self:
        """
        Centers the columns by subtracting each with its mean.

        Paramters
        ---------
        cols
            Any Polars expression that can be understood as columns.
        """
        self._steps.append(FitStep(partial(t.center), cols, self.exclude))
        return self

    def select(self, *cols: IntoExprColumn) -> Self:
        """
        Selects the columns from the dataset.

        Paramters
        ---------
        cols
            Any Polars expression that can be understood as columns.
        """
        self._steps.append(ExprStep(list(cols), PLContext.SELECT))
        return self
    
    def select_by_std(self, min_: float, max_: float) -> Self:
        """
        Fits and keeps columns that have standard deviation between `min_` and `max_`.
        Non-numeric columns will always be selected. Only numeric columns with std outside
        the `min_` `max_` bounds will be removed. This will never exclude the target if target
        is set.

        Parameters
        ----------
        min_
            Min standard deviation to select, inclusive.
        max_ 
            Max standard deviation to select, exclusive
        """
        self._steps.append(FitStep(partial(t.select_by_std, min_=min_, max_=max_), pl.col("*"), self.exclude, PLContext.SELECT))
        return self

    def polynomial_features(
        self, cols: List[str], degree: int, interaction_only: bool = True
    ) -> Self:
        """
        Generates polynomial combinations out of the features given, at the given degree.

        Parameters
        ----------
        cols
            A list of strings representing column names. Input to this function cannot be Polars expressions.
        degree
            The degree of the polynomial combination
        interaction_only
            It true, only combinations that involve 2 or more variables will be used.
        """
        if not all(isinstance(s, str) for s in cols):
            raise ValueError(
                "Input columns to `polynomial_features` must all be strings represeting column names."
            )

        self._steps.append(
            ExprStep(
                t.polynomial_features(cols, degree=degree, interaction_only=interaction_only),
                PLContext.WITH_COLUMNS,
            )
        )
        return self

    def winsorize(
        self,
        cols: IntoExprColumn,
        q_low: float = 0.05,
        q_high: float = 0.95,
        method: QuantileMethod = "nearest",
    ) -> Self:
        """
        Learns the lower and upper percentile from the columns, then clip each end at those values.
        If you wish to clip by constant values, you may append expression like pl.col(c).clip(c1, c2),
        where c1 and c2 are constants decided by the user.

        Parameters
        ----------
        cols
            Any Polars expression that can be understood as columns. Columns must be numerical.
        q_low
            The lower quantile value
        q_high
            The higher quantile value
        method
            Method to compute quantile. One of `nearest`, `higher`, `lower`, `midpoint`, `linear`.
        """
        self._steps.append(
            FitStep(
                partial(t.winsorize, q_low=q_low, q_high=q_high, method=method),
                cols,
                self.exclude,
            )
        )
        return self

    def drop(self, cols: IntoExprColumn) -> Self:
        """
        Drops the columns from the dataset.

        Paramters
        ---------
        cols
            Any Polars expression that can be understood as columns.
        """
        self._steps.append(ExprStep(pl.exclude(cols), PLContext.SELECT))
        return self

    def rename(self, rename_dict: Dict[str, str]) -> Self:
        """
        Renames the columns by the mapping.

        Paramters
        ---------
        rename_dict
            The name mapping
        """
        old = list(rename_dict.keys())
        self._steps.append(
            ExprStep([pl.col(k).alias(v) for k, v in rename_dict.items()], PLContext.WITH_COLUMNS)
        )

        return self.drop([o for o in old if o not in set(rename_dict.values())])

    def one_hot_encode(
        self,
        cols: IntoExprColumn | None = None,
        separator: str = "_",
        drop_first: bool = False,
        drop_cols: bool = True,
    ) -> Self:
        """
        Find the unique values in the string/categorical columns and one-hot encode them. This will NOT
        consider nulls as one of the unique values. Append pl.col(c).is_null().cast(pl.UInt8)
        expression to the pipeline if you want null indicators.

        If `cols` is None, this will pick all string and categorical columns.

        Parameters
        ----------
        cols
            Any Polars expression that can be understood as columns. If None, all string/categorical columns will be encoded.
        separator
            E.g. if column name is `col` and `a` is an elemenet in it, then the one-hot encoded column will be called
            `col_a` where the separator `_` is used.
        drop_first
            Whether to drop the first distinct value (in terms of str/categorical order). This helps with reducing
            dimension and prevents some issues from linear dependency.
        drop_cols
            Whether to drop the original columns after the transform
        """
        # First append new columns
        self._steps.append(
            FitStep(
                partial(t.one_hot_encode, separator=separator, drop_first=drop_first),
                cols if cols is not None else cs.string() | cs.categorical(),
                self.exclude,
            )
        )
        # Whether to drop the og str columns
        if drop_cols:
            if cols is None:
                return self.drop([pl.String, pl.Categorical])
            return self.drop(cols)
        return self

    def ordinal_encode(
        self,
        cols: IntoExprColumn | None = None,
        unknown_value: float | None = None,
        null_value: float | None = None,
    ) -> Self:
        """
        Find the unique values in the string/categorical columns and one-hot encode them. This will NOT
        consider nulls as one of the unique values. Append pl.col(c).is_null().cast(pl.UInt8)
        expression to the pipeline if you want null indicators.

        If `cols` is None, this will pick all string and categorical columns.

        Parameters
        ----------
        df
            Either a lazy or an eager dataframe
        cols
            A list of strings representing column names.
        unknown_value
            What to assign to values not seen in the initial dataset used to fit. None means null.
        null_value
            What to assign to values that are null in dataset.
        """
        self._steps.append(
            FitStep(
                partial(t.ordinal_encode, unknown_value=unknown_value, null_value=null_value),
                cols if cols is not None else cs.string() | cs.categorical(),
                self.exclude,
            )
        )
        return self

    def rank_hot_encode(
        self,
        col: str | pl.Expr,
        ranking: List[str],
        default_rank: int | None = None,
        drop_cols: bool = True,
    ) -> Self:
        """
        Given a ranking, e.g. ["bad", "neutral", "good"], where "bad", "neutral" and "good" are values coming
        from the column `col`, this will return two new columns, the first is ">=neutral", which
        will be 1 for all values in ("neutral", "good") and 0 otherwise, and the second new column is ">=good", which
        will be 1 for all values in ("good") and 0 otherwise.

        Parameters
        ----------
        col
            The name of a single column
        ranking
            A list of string representing the ranking of the values
        unknown_value
            What to assign to values not present in training dataset. None means null. Must be int.
        default_rank
            Default rank for all null/unseen values
        """
        self._steps.append(
            ExprStep(
                t.rank_hot_encode(col=col, ranking=ranking, default_rank=default_rank),
                PLContext.WITH_COLUMNS,
            )
        )
        if drop_cols:
            return self.drop(cols=[col])
        return self

    def target_encode(
        self,
        cols: IntoExprColumn | None = None,
        target: str | pl.Expr | None = None,
        min_samples_leaf: int = 20,
        smoothing: float = 10.0,
        default: EncoderDefaultStrategy | float | None = None,
    ) -> Self:
        """
        Target encode the given variables.

        Note: nulls will be encoded as well.

        Parameters
        ----------
        cols
            Any Polars expression that can be understood as columns. Columns of type != string/categorical
            will not produce any expression. If None, all string/categorical columns will be used.
        target
            The target column
        min_samples_leaf
            A regularization factor
        smoothing
            Smoothing effect to balance categorical average vs prior
        default
            If a new value is encountered during transform (unseen in training dataset), it will be mapped to default.
            If this is a string, it can be `null`, `zero`, or `mean`, where `mean` means map them to the mean of the target.

        Reference
        ---------
        https://contrib.scikit-learn.org/category_encoders/targetencoder.html
        """
        self._steps.append(
            FitStep(
                partial(
                    t.target_encode,
                    target=self._get_target(target),
                    min_samples_leaf=min_samples_leaf,
                    smoothing=smoothing,
                    default=default,
                ),
                cols if cols is not None else cs.string() | cs.categorical(),
                self.exclude,
            )
        )
        return self

    def woe_encode(
        self,
        cols: IntoExprColumn | None = None,
        target: str | pl.Expr | None = None,
        default: EncoderDefaultStrategy | float | None = None,
    ) -> Self:
        """
        Use Weight of Evidence to encode a discrete variable x with respect to target. This assumes x
        is discrete and castable to String. A value of 1 is added to all events/non-events
        (goods/bads) to smooth the computation. This is -1 * output of the package category_encoder's WOEEncoder.

        Note: nulls will be encoded as well.

        Parameters
        ----------
        cols
            Any Polars expression that can be understood as columns. Columns of type != string/categorical
            will not produce any expression. If None, all string/categorical columns will be used.
        target
            The target column
        default
            If a new value is encountered during transform (unseen in training dataset), it will be mapped to default.
            If this is a string, it can be `null`, `zero`, or `mean`, where `mean` means map them to the mean of the target.

        Reference
        ---------
        https://www.listendata.com/2015/03/weight-of-evidence-woe-and-information.html
        """
        self._steps.append(
            FitStep(
                partial(
                    t.woe_encode,
                    target=self._get_target(target),
                    default=default,
                ),
                cols if cols is not None else cs.string() | cs.categorical(),
                self.exclude,
            )
        )
        return self

    def iv_encode(
        self,
        cols: IntoExprColumn | None = None,
        target: str | pl.Expr | None = None,
        default: EncoderDefaultStrategy | float | None = None,
    ) -> Self:
        """
        Use Information Value to encode a discrete variable x with respect to target. This assumes x
        is discrete and castable to String. A value of 1 is added to all events/non-events
        (goods/bads) to smooth the computation.

        Note: nulls will be encoded as well.

        Parameters
        ----------
        cols
            Any Polars expression that can be understood as columns. Columns of type != string/categorical
            will not produce any expression. If None, all string/categorical columns will be used.
        target
            The target column
        default
            If a new value is encountered during transform (unseen in training dataset), it will be mapped to default.
            If this is a string, it can be `null`, `zero`, or `mean`, where `mean` means map them to the mean of the target.

        Reference
        ---------
        https://www.listendata.com/2015/03/weight-of-evidence-woe-and-information.html
        """
        self._steps.append(
            FitStep(
                partial(
                    t.iv_encode,
                    target=self._get_target(target),
                    default=default,
                ),
                cols if cols is not None else cs.string() | cs.categorical(),
                self.exclude,
            )
        )
        return self

    def with_columns(self, *exprs: pl.Expr) -> Self:
        """
        Run Polars with_columns for the expressions.
        """
        self._steps.append(ExprStep(list(exprs), PLContext.WITH_COLUMNS))
        return self

    def sort(
        self, by: IntoExprColumn, descending: bool | List[bool], maintain_order: bool = True
    ) -> Self:
        """Sorts the dataframe by the columns.

        Parameters
        ----------
        by
            The columns to sort by
        descending
            Whether the sort should be descending for the corresponding sort column
        maintain_order
            Whether the sort should maintain order or not
        """
        self._steps.append(SortStep(by=by, descending=descending, maintain_order=maintain_order))
        return self

    def explode(self, columns: str | pl.Expr | List[str] | List[pl.Expr]) -> Self:
        """Transform that represents `df.explode(columns)`"""
        if isinstance(columns, (str, pl.Expr)):
            exprs = [to_expr(columns)]
        elif isinstance(columns, list):
            exprs = [to_expr(c) for c in columns]
        else:
            raise ValueError(
                "Input `columns` must be a string, or a pl.Expr or a list of str or pl.Expr."
            )

        self._steps.append(ExprStep(exprs, PLContext.EXPLODE))
        return self

    def group_by_agg(
        self, by: IntoExprColumn, agg: List[pl.Expr], maintain_order: bool = False
    ) -> Self:
        """
        Performs a group by and agg on the data.

        Parameters
        ----------
        by
            The columns to group by
        agg
            The aggregation functions to run
        maintain_order
            Whether to maintain the group by order
        """
        self._steps.append(
            GroupByAggStep(by=by, agg=[to_expr(a) for a in agg], maintain_order=maintain_order)
        )
        return self

    def group_by_dynamic_agg(
        self,
        index_column: str,
        agg: List[pl.Expr],
        every: str,
        period: str | None = None,
        offset: str | None = None,
        include_boundaries: bool = False,
        closed: Literal["left", "right", "both", "none"] = "left",
        label: Literal["left", "right", "datapoint"] = "left",
        group_by: IntoExprColumn | None = None,
        start_by: Literal[
            "window",
            "datapoint",
            "monday",
            "tuesday",
            "wednesday",
            "thursday",
            "friday",
            "saturday",
            "sunday",
        ] = "window",
    ) -> Self:
        """
        See polars group_by_dynamic documentation for an explanation on the input arguments.

        https://docs.pola.rs/api/python/stable/reference/dataframe/api/polars.DataFrame.group_by_dynamic.html#polars.DataFrame.group_by_dynamic
        """
        self._steps.append(
            GroupByDynAggStep(
                index_column=index_column,
                agg=[to_expr(a) for a in agg],
                every=every,
                group_by=group_by,
                period=period,
                offset=offset,
                include_boundaries=include_boundaries,
                closed=closed,
                label=label,
                start_by=start_by,
            )
        )
        return self

    # How to type this?
    def append_fit_func(self, func, cols: IntoExprColumn, **kwargs) -> Self:
        """
        Adds a custom transform that requires a fit step in the blueprint.

        Any custom function must satistfy the following function signature:
        my_func(df:Union[pl.DataFrame, pl.LazyFrame], cols: List[str], ...) -> List[pl.Expr]
        where ... means kwargs. The fit step "learns" the parameters needed to translate
        the transform into concrete expressions.

        Parameters
        ----------
        func
            A callable with signature (pl.DataFrame | pl.LazyFrame, cols: List[str], ...) -> ExprTransform,
        cols
            The columns to be fed into the func. Note that in func's signature, a list of strings
            should be expected. But here, cols can be any polars selector expression. The reason is that
            during "fit", cols is turned into concrete column names.
        **kwargs
            Any other arguments to func must be passed as kwargs
        """
        import inspect

        keywords = kwargs.copy()
        if "target" in inspect.signature(func).parameters:  # func has "target" as input
            if "target" not in kwargs:  # if target is not explicitly given
                keywords["target"] = self._get_target()
                if keywords["target"] is None:
                    raise ValueError(
                        "Target is not explicitly given and is required by the custom function."
                    )

        self._steps.append(
            FitStep(
                partial(func, **keywords),
                cols,
                self.exclude,
            )
        )
        return self

    def append_step_from_dict(self, dictionary: Dict[str, Any]) -> Self:
        """
        Append a step to the blueprint by taking in a dictionary with keys `name`, `args`, and `kwargs`, where
        the value of args must be a List[Any] and the value of kwargs must be a dict[str, Any].
        """
        step_repr: StepRepr = StepRepr.from_dict(dictionary)
        func = getattr(self, step_repr.name, None)  # Default is None
        if func is None or step_repr.name.startswith("_"):
            raise ValueError("Unknown or invalid method name.")

        return func(*step_repr.args, **step_repr.kwargs)

    def materialize(self, **kwargs) -> Pipeline:
        """
        Materialize the blueprint, which means that it will fit and learn all the paramters needed.

        All kwargs here will be passed to polars.LazyFrame.collect()
        """
        transforms: List[PipelineStep] = []
        df: pl.DataFrame = self._df.collect(**kwargs)
        # Let this lazy plan go through the fit process. The frame will be collected temporarily but
        # the collect should be and optimized.
        df_lazy: pl.LazyFrame = df.lazy()
        for step in self._steps:
            if isinstance(step, FitStep):  # Need fitting, which is done here
                df_temp: pl.DataFrame = df_lazy.collect(**kwargs)
                exprs = step.fit(df_temp)
                step = ExprStep(exprs, step.context)
                transforms.append(step)
                df_lazy = step.apply_df(df_temp.lazy())
            elif isinstance(step, tuple(_SERIALIZABLE_STEPS)):
                transforms.append(step)
                df_lazy = step.apply_df(df_lazy)
            else:
                raise ValueError(f"Not a valid step: {step.__class__}")

        return Pipeline(
            name=self.name,
            feature_names_in_=list(self.feature_names_in_),
            feature_names_out_=df_lazy.collect_schema().names(),
            transforms=transforms,
            lowercase=self.lowercase,
            uppercase=self.uppercase,
        )

    def fit(self, X=None, y=None, **kwargs) -> Pipeline:
        """
        Alias for self.materialize()
        """
        return self.materialize(**kwargs)

    def transform(self, df: PolarsFrame, **kwargs) -> pl.DataFrame:
        """
        Fits the blueprint with the dataframe that it is initialized with, and
        transforms the input dataframe.

        Parameters
        ----------
        df
            Any Polars dataframe
        **kwargs
            Will be passed to Pipeline's `transform` method.
        """
        return self.materialize().transform(df, **kwargs)



"""Not meant for outside use."""

from __future__ import annotations

import polars as pl
from typing import Sequence
from pathlib import Path
from polars.plugins import register_plugin_function

# Only need this
_PLUGIN_PATH = Path(__file__).parent

# V1.18 Introduces a Int128 dtype
# _IS_POLARS_V1_18 = pl.__version__.startswith("1.18.")


def pl_plugin(
    *,
    symbol: str,
    args: Sequence[pl.Series | pl.Expr],
    kwargs: dict[str, str | int | float | bool] | None = None,
    is_elementwise: bool = False,
    returns_scalar: bool = False,
    changes_length: bool = False,
    cast_to_supertype: bool = False,
    pass_name_to_apply: bool = False,
) -> pl.Expr:
    return register_plugin_function(
        plugin_path=_PLUGIN_PATH,
        args=args,
        function_name=symbol,
        kwargs=kwargs,
        is_elementwise=is_elementwise,
        returns_scalar=returns_scalar,
        changes_length=changes_length,
        cast_to_supertype=cast_to_supertype,
        pass_name_to_apply=pass_name_to_apply,
    )


# Auxiliary functions for type conversions
def to_expr(e: str | pl.Expr | int | float) -> pl.Expr:
    """
    Turns a string into an expression

    Parameters
    ----------
    e
        Either a str represeting a column name or an expression
    """
    if isinstance(e, str):
        return pl.col(e)
    elif isinstance(e, (int, float)):
        return pl.lit(e)
    elif isinstance(e, pl.Expr):
        return e
    else:
        raise ValueError("Input must either be a string or a Polars expression.")



from __future__ import annotations
from typing import Literal, List, Callable, Union
import sys
import polars as pl

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:  # 3.9, 3.8
    from typing_extensions import TypeAlias

# Custom "Enum" Types
DetrendMethod: TypeAlias = Literal["linear", "mean"]
Alternative: TypeAlias = Literal["two-sided", "less", "greater"]
Distance: TypeAlias = Literal["l1", "l2", "sql2", "inf", "cosine", "haversine"]
KdtDistance: TypeAlias = Literal["l1", "l2", "sql2", "inf"]
ConvMode: TypeAlias = Literal["same", "left", "right", "full", "valid"]
ConvMethod: TypeAlias = Literal["fft", "direct"]
CorrMethod: TypeAlias = Literal["pearson", "spearman", "xi", "kendall", "bicor"]
SimpleImputeMethod: TypeAlias = Literal["mean", "median", "mode"]
SimpleScaleMethod: TypeAlias = Literal["min_max", "standard", "abs_max"]
Noise: TypeAlias = Literal["gaussian", "uniform"]
LRMethods: TypeAlias = Literal["normal", "l2", "l1"]
LRSolverMethods: TypeAlias = Literal["svd", "qr", "cholesky"]
NullPolicy: TypeAlias = Literal["raise", "skip", "one", "zero", "ignore"]
MultiAUCStrategy: TypeAlias = Literal["weighted", "macro"]
EncoderDefaultStrategy: TypeAlias = Literal["mean", "null", "zero"]
# Copy of Polars
QuantileMethod: TypeAlias = Literal["nearest", "higher", "lower", "midpoint", "linear"]

# Other Custom Types
PolarsFrame: TypeAlias = Union[pl.DataFrame, pl.LazyFrame]
ExprTransform: TypeAlias = Union[pl.Expr, List[pl.Expr]]
# Need ...
FitTransformFunc: TypeAlias = Callable[[PolarsFrame, List[str]], ExprTransform]

# # For compatibility
# IntoNumpy: TypeAlias = Union["Sequence[float]", "Sequence[int]"]
# """Anything which can be converted to a NumPy numeric array.

# Examples:
#     >>> from polars_ds.typing import IntoNumPy
#     >>> def agnostic_to_numpy(s: IntoNumpy) -> np.ndarray:
#     ...     return s.to_numpy()
# """



from __future__ import annotations

import polars as pl
import random
import math
from typing import List, Tuple, Dict
from itertools import combinations, islice

# Internal dependency
from polars_ds.typing import PolarsFrame

__all__ = ["sample", "volume_neutral", "downsample", "random_cols", "split_by_ratio"]


def _sampler_expr(value: float | int, seed: int | None = None) -> pl.Expr:
    if isinstance(value, float):
        if value >= 1.0 or value <= 0.0:
            raise ValueError("Sample rate must be in (0, 1) range.")
        return pl.int_range(0, pl.len(), dtype=pl.UInt32).shuffle(seed) < (pl.len() * value).cast(
            pl.UInt32
        )
    elif isinstance(value, int):
        if value <= 0:
            raise ValueError("Sample count must be > 0.")
        return pl.int_range(0, pl.len(), dtype=pl.UInt32).shuffle(seed) < pl.lit(
            value, dtype=pl.UInt32
        )
    elif isinstance(value, pl.Expr):
        return NotImplemented
    else:
        raise ValueError("Sample value must be either int or float.")


def sample(df: PolarsFrame, value: float | int, seed: int | None = None) -> pl.DataFrame:
    """
    Samples the dataframe.

    Parameters
    ----------
    df
        Either a lazy or eager Polars dataframe
    value
        Either an int, in which case it means get n random rows from the df, or a float in the range (0, 1), in which
        case it means sample x% of the dataframe.
    seed
        A random seed
    """
    return df.lazy().filter(_sampler_expr(value, seed)).collect()


def volume_neutral(
    df: PolarsFrame,
    by: pl.Expr,
    control: pl.Expr | List[pl.Expr] | None = None,
    target_volume: int | None = None,
    seed: int | None = None,
) -> pl.DataFrame:
    """
    Select volume neutral many population from each segment in `by`, with optional control categories.

    Parameters
    ----------
    df
        Either a lazy or eager Polars dataframe
    by
        A Polars expression represeting a column with discrete values. Volume neutral by the values
        in this column.
    control
        Additional level(s). If not none, the volume neutral selection will happen at the
        sublevel of the control column(s). See example.
    target_volume
        If none, it will select min(a, b, c) rows.
    seed
        A random seed
    """
    if isinstance(target_volume, int):
        target = pl.min_horizontal(by.value_counts().struct.field("count").min(), target_volume)
    else:
        target = by.value_counts().struct.field("count").min()

    if isinstance(control, pl.Expr):
        ctrl = [control]
    elif isinstance(control, list):
        ctrl = [c for c in control if isinstance(c, pl.Expr)]
    else:
        ctrl = []

    if len(ctrl) > 0:
        target = target.over(ctrl)
        final_ref = ctrl + [by]
    else:
        final_ref = by

    return (
        df.lazy().filter(pl.int_range(0, pl.len()).shuffle(seed).over(final_ref) < target).collect()
    )


def downsample(
    df: PolarsFrame,
    *conditions: Tuple[pl.Expr, float | int],
    seed: int | None = None,
) -> pl.DataFrame:
    """
    Downsamples data on the subsets where condition is true.

    Parameters
    ----------
    df
        Either a lazy or eager Polars dataframe
    conditions
        Tuple[pl.Expr, float|int] or a sequence of such tuples as positional arguments.
        The first entry in the tuple should be a boolean expression and the second entry means we sample
        either n or x% on the part where the boolean is true.
    seed
        A random seed

    Example
    -------
    >>> import polars_ds.sample as sa
    >>> df.group_by("y").len().sort("y")
    shape: (4, 2)
    ┌─────┬───────┐
    │ y   ┆ len   │
    │ --- ┆ ---   │
    │ i32 ┆ u32   │
    ╞═════╪═══════╡
    │ 0   ┆ 24875 │
    │ 1   ┆ 25172 │
    │ 2   ┆ 25018 │
    │ 3   ┆ 24935 │
    └─────┴───────┘
    >>> downsampled = sa.down_sample(
    >>>     df,
    >>>     [(pl.col("y") == 1, 0.5), (pl.col("y") == 2, 0.5)]
    >>> )
    >>> downsampled.group_by("y").len().sort("y")
    shape: (4, 2)
    ┌─────┬───────┐
    │ y   ┆ len   │
    │ --- ┆ ---   │
    │ i32 ┆ u32   │
    ╞═════╪═══════╡
    │ 0   ┆ 24875 │
    │ 1   ┆ 12586 │
    │ 2   ┆ 12509 │
    │ 3   ┆ 24935 │
    └─────┴───────┘
    >>> # And this is equivalent to
    >>> downsampled = sa.down_sample(
    >>>     df,
    >>>     [(pl.col("y").is_between(1, 2, closed="both"), 0.5)]
    >>> )
    """
    all_filters = ((_sampler_expr(r, seed).over(c) | (~c)) for c, r in conditions)
    return df.lazy().filter(pl.lit(True).and_(*all_filters)).collect()


def random_cols(
    all_columns: List[str],
    k: int,
    keep: List[str] | None = None,
    seed: int | None = None,
) -> List[str]:
    """
    Selects random columns from the given pool of columns. Returns the selected columns in a list.
    Note, it is impossible for this to randomly select both ["x", "y"] and ["y", "x"].

    Parameters
    ----------
    all_columns
        All column names
    k
        Select k random columns from all columns outside of `keep`.
    keep
        Columns to always keep
    seed
        A random seed
    """
    if keep is None:
        out = []
        to_sample = combinations(all_columns, k)
    else:
        out = keep
        to_sample = combinations((c for c in all_columns if c not in keep), k)

    pool_size = len(all_columns) - len(out)
    if pool_size < k:
        raise ValueError("Not enough columns to select from.")

    n = random.randrange(0, math.comb(pool_size, k))
    rand_cols = next(islice(to_sample, n, None), None)
    return out + list(rand_cols)


def split_by_ratio(
    df: PolarsFrame,
    split_ratio: float | List[float] | Dict[str, float],
    seed: int | None = None,
    split_col: str = "__split",
    by: str | list[str] | None = None,
    default_split_1: str = "train",
    default_split_2: str = "test",
) -> PolarsFrame:
    """
    Creates a random `split` column in the dataframe.

    If split_ratio is a single number, it is treated as the ratio for the default_split_1 set, which by
    default has the name 'train', and (1-ratio) is always the ratio for the default_split_2 set, which by
    default has the name 'test'.

    If split_ratio is a list of floats, then they must sum to 1 and the return will be the corresponding ratios
    and the split column values will be "split_i".

    If the split_ratio is a dict, then the dict values will be the ratios and the dict keys will be the value
    in the split column.

    Please avoid using floating point values with too many decimal places, which may cause
    the splits to be off by a 1 row.

    This will return lazy frames if input is lazy, and eager frames if input is eager. But if `by` is given,
    then the dataframe will be collected and will only return eager dataframe in the end.

    Parameters
    ----------
    df
        Either a lazy or eager Polars dataframe
    split_ratio
        Either a single float or a list of floats.
    split_col
        The column name of the split
    seed
        The random seed
    by
        Optional str of list of str. The column(s) to stratify by.
    default_split_1
        Name of the first split when split_ratio is a scalar
    default_split_2
        Name of the second split when split_ratio is a scalar
    """

    if by is not None:
        results = [
            split_by_ratio(
                frame,
                split_ratio=split_ratio,
                seed=seed,
                by=None,
                split_col=split_col,
                default_split_1=default_split_1,
                default_split_2=default_split_2,
            )
            for frame in df.lazy().collect().partition_by(by)
        ]
        return pl.concat(results, how="vertical")

    if isinstance(split_ratio, float):
        if split_ratio <= 0.0 or split_ratio >= 1:
            raise ValueError("Split ratio must be > 0 and < 1.")

        return (
            df.with_row_index(name="__id")
            .with_columns(
                pl.when(pl.col("__id").shuffle(seed=seed) < (pl.len() * split_ratio).cast(pl.Int64))
                .then(pl.lit(default_split_1, dtype=pl.String))
                .otherwise(pl.lit(default_split_2, dtype=pl.String))
                .alias(split_col)
            )
            .select(pl.all().exclude("__id"))
        )

    else:
        if len(split_ratio) == 1:
            raise ValueError("If split_ratio is not a scalar, it must have length > 1.")
        else:
            if isinstance(split_ratio, dict):
                ratios: pl.Series = pl.Series(split_ratio.values())
                split_names = [str(k) for k in split_ratio.keys()]
            else:  # should work with other iterables like tuple
                ratios: pl.Series = pl.Series(split_ratio)
                split_names = [f"split_{i}" for i in range(len(split_ratio))]

            if ratios.sum() != 1:
                raise ValueError("Sum of the ratios is not 1.")

            pct = ratios.cum_sum()
            expr = pl.when(pl.lit(False)).then(None)
            for p, k in zip(pct, split_names):
                expr = expr.when(pl.col("__pct") < p).then(pl.lit(k, dtype=pl.String))

            return (
                df.with_row_index(name="__id")
                .with_columns(pl.col("__id").shuffle(seed=seed).alias("__tt"))
                .sort("__tt")
                .with_columns((pl.col("__tt") / pl.len()).alias("__pct"))
                .select(expr.alias(split_col), pl.all().exclude(["__id", "__pct", "__tt"]))
            )



# from __future__ import annotations
# from typing import Iterable
# import polars as pl
# from .type_alias import StrOrExpr, str_to_expr, Distance
# from ._utils import pl_plugin


# def query_bt_knn_ptwise(
#     *features: str | pl.Expr,
#     index: str | pl.Expr,
#     r: float,
#     distance_metric: Distance = "euclidean",
#     sort: bool = True,
#     parallel: bool = False,
#     k: int = None,
#     return_dist: bool = False,
# ) -> pl.Expr:
#     """
#     Takes an index column, uses the feature columns to determine distance and finds all neighbors
#     within distance r from each id. It returns a list containing (neighbor_id, distance) tuples.
#     This is bounded by max k neighbors.
#     index columns must be convertible to u32

#     Parameters
#     ----------
#     *features : str | pl.Expr
#         The feature columns.
#     index : str | pl.Expr
#         The index column.
#     r : float
#         The radius.
#     distance_metric : Distance, optional
#         The distance metric to use, by default "euclidean". Currently the only available options are "euclidean" and "haversine".
#     sort : bool, optional
#         Whether to sort the output, by default True.
#     parallel : bool, optional
#         Whether to evaluate this in parallel, by default False.
#     k : int, optional
#     """
#     if r < 0:
#         raise ValueError("r must be positive")
#     elif isinstance(r, pl.Expr):
#         raise ValueError(
#             "r must be a scalar float. Expressions are not supported.")

#     if distance_metric.lower() not in ("euclidean", "haversine"):
#         raise ValueError(
#             "Invalid distance metric. Must be 'euclidean' or 'haversine'.")
#     if len(features) == 0:
#         raise ValueError("Must provide at least one feature column.")
#     elif len(features) != 2 and distance_metric == "haversine":
#         raise ValueError("Haversine distance requires exactly 2 features.")

#     idx = str_to_expr(index).cast(pl.UInt32).rechunk()

#     if not k:
#         # the length of any of the feature columns
#         k = pl.col(features[0]).len()
#     if k < 1:
#         raise ValueError("k must be positive.")

#     # Columns to send over to Rust as &[Series]
#     # First column will always be the index column
#     cols = [idx]
#     cols.extend([str_to_expr(f) for f in features])
#     if return_dist:
#         return pl_plugin(
#             symbol="pl_query_knn_ptwise_wdist",
#             args=cols,
#             kwargs={"r": r, "distance_metric": distance_metric.lower(
#             ), "sort": sort, "parallel": parallel, "k": k},
#             is_elementwise=False,
#         )
#     else:
#         return pl_plugin(
#             symbol="pl_query_knn_ptwise",
#             args=cols,
#             kwargs={"r": r, "distance_metric": distance_metric.lower(
#             ), "sort": sort, "parallel": parallel, "k": k},
#             is_elementwise=False,
#         )


# def query_bt_knn_radius_freq_cnt(
#     *features: str | pl.Expr,
#     index: str | pl.Expr,
#     r: float = None,
#     distance_metric: Distance = "euclidean",
#     sort: bool = True,
#     parallel: bool = False,
#     k: int = None,
# ) -> pl.Expr:
#     """
#     Returns the frequency count of neighbors within distance r, and within k nearest neighbors for each row using the given distance metric.
#     The point is always a neighbor of itself.
#     We need an index column that can be cast to u32.

#     Parameters
#     ----------

#     *features : str | pl.Expr
#         The feature columns.
#     index : str | pl.Expr
#         The index column.
#     r : float, optional
#         The radius. If None, it will be set to infinity, by default None.
#     distance_metric : Distance, optional
#         The distance metric to use, by default "euclidean". Currently the only available options are "euclidean" and "haversine".
#     sort : bool, optional
#         Whether to sort the output, by default True.
#     parallel : bool, optional
#         Whether to evaluate this in parallel, by default False.
#     k : int, optional
#         The number of nearest neighbors to consider, by default 1.
#     """
#     if not k:
#         # the length of any of the feature columns
#         k = pl.col(features[0]).len()
#     if not r:
#         r = float("inf")
#     if r < 0:
#         raise ValueError("r must be positive")
#     elif isinstance(r, pl.Expr):
#         raise ValueError(
#             "r must be a scalar float. Expressions are not supported.")

#     if distance_metric.lower() not in ("euclidean", "haversine"):
#         raise ValueError(
#             "Invalid distance metric. Must be 'euclidean' or 'haversine'.")
#     if len(features) == 0:
#         raise ValueError("Must provide at least one feature column.")
#     elif len(features) != 2 and distance_metric == "haversine":
#         raise ValueError("Haversine distance requires exactly 2 features.")

#     if k < 1:
#         raise ValueError("k must be positive.")
#     idx = str_to_expr(index).cast(pl.UInt32).rechunk()
#     # Columns to send over to Rust as &[Series]
#     # First column will always be the index column
#     cols = [idx]
#     cols.extend([str_to_expr(f) for f in features])
#     knn_expr: pl.Expr = query_bt_knn_ptwise(
#         *features, index=index, r=r, distance_metric=distance_metric, sort=sort, parallel=parallel, k=k, return_dist=False)
#     return knn_expr.explode().drop_nulls().value_counts(sort=True, parallel=parallel)


# def query_bt_knn_avg(
#     *features: str | pl.Expr,
#     index: str | pl.Expr,
#     r: float,
#     distance_metric: Distance = "euclidean",
#     sort: bool = True,
#     parallel: bool = False,
#     k: int = None,
# ) -> pl.Expr:
#     """Takes an index column, uses the feature columns to determine distance and finds all neighbors
#     within distance r from each id. It returns a list containing (neighbor_id, distance) tuples.
#     index columns must be convertible to u32

#     Args:
#         index (str | pl.Expr): _description_
#         r (float): _description_
#         dist (Distance, optional): _description_. Defaults to "euclidian".
#         sort (bool, optional): _description_. Defaults to True.
#         parallel (bool, optional): _description_. Defaults to False.

#     Returns:
#         pl.Expr: _description_
#     """
#     if not k:
#         # the length of any of the feature columns
#         k = pl.col(features[0]).len()
#     if r < 0:
#         raise ValueError("r must be positive")
#     elif isinstance(r, pl.Expr):
#         raise ValueError(
#             "r must be a scalar float. Expressions are not supported.")

#     if distance_metric.lower() not in ("euclidean", "haversine"):
#         raise ValueError(
#             "Invalid distance metric. Must be 'euclidean' or 'haversine'.")
#     if len(features) == 0:
#         raise ValueError("Must provide at least one feature column.")
#     elif len(features) != 2 and distance_metric == "haversine":
#         raise ValueError("Haversine distance requires exactly 2 features.")

#     if k < 1:
#         raise ValueError("k must be positive.")
#     idx = str_to_expr(index).cast(pl.UInt32).rechunk()
#     # Columns to send over to Rust as &[Series]
#     # First column will always be the index column
#     cols = [idx]
#     cols.extend([str_to_expr(f) for f in features])
#     return pl_plugin(
#         symbol="pl_ball_tree_knn_avg",
#         args=cols,
#         kwargs={"r": r, "distance_metric": distance_metric.lower(
#         ), "sort": sort, "parallel": parallel, "k": k},
#         is_elementwise=False,
#     )


# def bt_within_dist_from(
#     *features: str | pl.Expr,
#     pt: Iterable[float],
#     r: float | str | pl.Expr,
#     distance_metric: Distance = "euclidean",
#     parallel: bool = False,
# ) -> pl.Expr:
#     """
#     Returns a boolean column indicating if the provided point is within distance r.

#     Parameters
#     ----------
#     *features : str | pl.Expr
#         The feature columns.
#     pt : Iterable[float]
#         The point to compare against.
#     r : float | str | pl.Expr
#         The radius. Either a scalar float, or a 1d array with len = row_count(X).
#     distance_metric : Distance, optional
#         The distance metric to use, by default "euclidean". Currently the only available
#         options are "euclidean" and "haversine".
#     """
#     if distance_metric.lower() not in ("euclidean", "haversine"):
#         raise ValueError(
#             "Invalid distance metric. Must be 'euclidean' or 'haversine'.")
#     if len(features) < 2:
#         raise ValueError("Must provide at least two feature columns.")
#     if len(pt) != len(features):
#         raise ValueError(
#             "Number of features must match the number of dimensions")

#     if isinstance(r, (float, int)):
#         rad = pl.lit(pl.Series(values=[r], dtype=pl.Float64))
#     elif isinstance(r, pl.Expr):
#         rad = r
#     elif isinstance(r, str):
#         rad = pl.col(r)
#     else:
#         rad = pl.lit(pl.Series(values=r, dtype=pl.Float64))
#     cols = [rad]
#     cols.extend([str_to_expr(f) for f in features])
#     return pl_plugin(
#         symbol="pl_bt_within_dist_from",
#         args=cols,
#         kwargs={"point": pt, "distance_metric": distance_metric.lower(),
#                 "parallel": parallel},
#     )


# def is_bt_knn_from(
#     *features,
#     pt: Iterable[float],
#     k: int,
#     distance_metric: Distance = "euclidean",
#     parallel: bool = False,
#     epsilon: float = None,
# ) -> pl.Expr:
#     """
#     Returns a boolean column indicating if the provided point is within the k-nearest neighbors.

#     Parameters
#     ----------
#     *features : str | pl.Expr
#         The feature columns.
#     pt : Iterable[float]
#         The point to compare against.
#     k : int
#         Number of neighbors to consider.
#     distance_metric : Distance, optional
#         The distance metric to use, by default "euclidean". Currently the only available
#         options are "euclidean" and "haversine".
#     parallel : bool, optional
#     """
#     if distance_metric.lower() not in ("euclidean", "haversine"):
#         raise ValueError(
#             "Invalid distance metric. Must be 'euclidean' or 'haversine'.")
#     if len(features) < 2:
#         raise ValueError("Must provide at least two feature columns.")
#     if len(pt) != len(features):
#         raise ValueError(
#             "Number of features must match the number of dimensions")
#     if k < 1:
#         raise ValueError("k must be positive.")
#     cols = [str_to_expr(f) for f in features]
#     return pl_plugin(
#         symbol="pl_bt_knn_from",
#         args=cols,
#         kwargs={"point": pt, "distance_metric": distance_metric.lower(
#         ), "parallel": parallel, "k": k, "epsilon": epsilon},
#     )


# def query_bt_nb_cnt(
#     *features: str | pl.Expr,
#     r: float | str | Iterable[float],
#     index: str | pl.Expr,
#     distance_metric: Distance = "euclidean",
#     parallel: bool = False
# ) -> pl.Expr:
#     """
#     Returns the number of neighbors within ( <= ) radius r for each row using the given distance metric.
#     The point is always a neighbor of itself.

#     Parameters
#     ----------
#     r : float | str | Iterable[float]
#         The radius. Either a scalar float, or a 1d array with len = row_count(X).
#     index : str | pl.Expr
#         The index column.
#     *features : str | pl.Expr
#         The feature columns.
#     distance_metric : Distance, optional
#         The distance metric to use, by default "euclidean". Currently the only available
#         options are "euclidean" and "haversine".
#     parallel : bool, optional
#         Whether to evaluate this in parallel, by default False
#     """
#     if distance_metric.lower() not in ("euclidean", "haversine"):
#         raise ValueError(
#             "Invalid distance metric. Must be 'euclidean' or 'haversine'.")
#     if len(features) == 0:
#         raise ValueError("Must provide at least one feature column.")
#     elif len(features) != 2 and distance_metric == "haversine":
#         raise ValueError("Haversine distance requires exactly 2 features.")

#     if isinstance(r, (float, int)):
#         rad = pl.lit(pl.Series(values=[r], dtype=pl.Float64))
#     elif isinstance(r, pl.Expr):
#         rad = r
#     elif isinstance(r, str):
#         rad = pl.col(r)
#     else:
#         rad = pl.lit(pl.Series(values=r, dtype=pl.Float64))

#     idx = str_to_expr(index).cast(pl.UInt32).rechunk()
#     cols = [idx, rad]
#     cols.extend([str_to_expr(f) for f in features])
#     return pl_plugin(
#         symbol="pl_nb_count",
#         args=cols,
#         kwargs={
#             "distance_metric": distance_metric.lower(),
#             "parallel": parallel,
#             # Stubbed values so we can use the
#             # same kwargs struct in rust
#             "sort": True,
#             "r": 0,
#             "k": 1,
#         },
#     )


# def query_bt_radius_ptwise(
#     *features: str | pl.Expr,
#     index: str | pl.Expr,
#     r: float,
#     distance_metric: Distance = "euclidean",
#     sort: bool = True,
#     parallel: bool = False,
#     k=None,
#     return_dist: bool = False,
# ) -> pl.Expr:
#     """
#     Returns a list of neighbors within distance r from each id.
#     index columns must be convertible to u32

#     Parameters
#     ----------
#     *features : str | pl.Expr
#         The feature columns.
#     index : str | pl.Expr
#         The index column.
#     r : float
#         The radius.
#     distance_metric : Distance, optional
#         The distance metric to use, by default "euclidean". Currently the only available options are "euclidean" and "haversine".
#     sort : bool, optional
#         Whether to sort the output, by default True.
#     parallel : bool, optional
#         Whether to evaluate this in parallel, by default False.
#     k : int, optional
#         Max number of neighbors to consider.
#     """
#     return query_bt_knn_ptwise(*features, index=index, r=r, distance_metric=distance_metric, sort=sort, parallel=parallel, k=k, return_dist=return_dist)


# def query_bt_radius_freq_cnt(
#     *features: str | pl.Expr,
#     index: str | pl.Expr,
#     r: float = None,
#     distance_metric: Distance = "euclidean",
#     sort: bool = True,
#     parallel: bool = False,
#     k: int = None,
# ) -> pl.Expr:
#     """
#     Returns the frequency count of neighbors within distance r, and within k nearest neighbors for each row using the given distance metric.
#     The point is always a neighbor of itself.
#     We need an index column that can be cast to u32.

#     Parameters
#     ----------

#     *features : str | pl.Expr
#         The feature columns.
#     index : str | pl.Expr
#         The index column.
#     r : float, optional
#         The radius. If None, it will be set to infinity, by default None.
#     distance_metric : Distance, optional
#         The distance metric to use, by default "euclidean". Currently the only available options are "euclidean" and "haversine".
#     sort : bool, optional
#         Whether to sort the output, by default True.
#     parallel : bool, optional
#         Whether to evaluate this in parallel, by default False.
#     k : int, optional
#         The number of nearest neighbors to consider, by default 1.
#     """
#     return query_bt_knn_radius_freq_cnt(*features, index=index, r=r, distance_metric=distance_metric, sort=sort, parallel=parallel, k=k)



"""Simple Statistics."""

from __future__ import annotations

import polars as pl
import math

# Internal dependencies
from polars_ds.typing import Alternative, CorrMethod, Noise, QuantileMethod
from polars_ds._utils import pl_plugin, to_expr

__all__ = [
    "ttest_ind",
    "ttest_1samp",
    "ttest_ind_from_stats",
    "ks_2samp",
    "f_test",
    "mann_whitney_u",
    "chi2",
    "perturb",
    "jitter",
    "add_noise",
    "normal_test",
    "random",
    "random_null",
    "random_int",
    "random_str",
    "random_binomial",
    "random_exp",
    "random_normal",
    "hmean",
    "gmean",
    "weighted_hmean",
    "weighted_gmean",
    "weighted_mean",
    "weighted_var",
    "weighted_cov",
    "weighted_corr",
    "cosine_sim",
    "weighted_cosine_sim",
    "xi_corr",
    "kendall_tau",
    "bicor",
    "corr",
]


def ttest_ind(
    var1: str | pl.Expr,
    var2: str | pl.Expr,
    alternative: Alternative = "two-sided",
    equal_var: bool = False,
) -> pl.Expr:
    """
    Performs 2 sample student's t test or Welch's t test. Functionality-wise this is desgined
    to be equivalent to SciPy's ttest_ind, with fewer options. The result is not exact but
    within 1e-10 precision from SciPy's.

    In the case of student's t test, the data is assumed to have no nulls, and n = expr.count()
    is used. Note expr.count() only counts non-null elements after polars 0.20.
    The degree of freedom will be 2n - 2. As a result, nulls might cause problems.

    In the case of Welch's t test, data will be sanitized (nulls, NaNs, Infs will be dropped
    before the test), and df will be counted based on the length of sanitized data.

    If (NaN, NaN) is returned, then it is possible that one of the following numeric
    problems occurred:

    1. There is no valid value in the inputs, or the mean is inf.
    2. Input variable has length 0 after removing non-finite values.

    Parameters
    ----------
    var1
        Variable 1
    var2
        Variable 2
    alternative : {"two-sided", "less", "greater"}
        Alternative of the hypothesis test
    equal_var
        If true, perform standard student t 2 sample test. Otherwise, perform Welch's
        t test.

    Examples
    --------
    Same length, equal variance comparisons.
    >>> df.select(pds.ttest_ind("x1", "x2", equal_var=True))

    Potentially unequal length, unequal variance.
    >>> df.select(
    ...     pds.ttest_ind(
    ...         pl.col("x1").filter(condition_A), pl.col("x1").filter(condition_B), equal_var=False
    ...     )
    ... )
    """
    y1, y2 = to_expr(var1), to_expr(var2)
    if equal_var:
        m1 = y1.mean()
        m2 = y2.mean()
        v1 = y1.var()
        v2 = y2.var()
        cnt = y1.count().cast(pl.UInt64)
        return pl_plugin(
            symbol="pl_ttest_2samp",
            args=[m1, m2, v1, v2, cnt, pl.lit(alternative, dtype=pl.String)],
            returns_scalar=True,
        )
    else:
        s1 = y1.filter(y1.is_finite())
        s2 = y2.filter(y2.is_finite())
        m1 = s1.mean()
        m2 = s2.mean()
        v1 = s1.var()
        v2 = s2.var()
        n1 = s1.len().cast(pl.UInt64)
        n2 = s2.len().cast(pl.UInt64)
        return pl_plugin(
            symbol="pl_welch_t",
            args=[m1, m2, v1, v2, n1, n2, pl.lit(alternative, dtype=pl.String)],
            returns_scalar=True,
        )


def ttest_1samp(
    var1: str | pl.Expr, pop_mean: float, alternative: Alternative = "two-sided"
) -> pl.Expr:
    """
    Performs a standard 1 sample t test using reference column and expected mean. This function
    sanitizes the self column first. The df is the count of valid values.

    If (NaN, NaN) is returned, then it is possible that one of the following numeric
    problems occurred:

    1. There is no valid value in the inputs, or the mean is inf.
    2. Input variable has length 0 after removing non-finite values.

    Parameters
    ----------
    var1
        Variable 1
    pop_mean
        The expected population mean in the hypothesis test
    alternative : {"two-sided", "less", "greater"}
        Alternative of the hypothesis test
    """
    y = to_expr(var1)
    s1 = y.filter(y.is_finite())
    sm = s1.mean()
    pm = pl.lit(pop_mean, dtype=pl.Float64)
    var = s1.var()
    cnt = s1.len().cast(pl.UInt64)
    alt = pl.lit(alternative, dtype=pl.String)
    return pl_plugin(
        symbol="pl_ttest_1samp",
        args=[sm, pm, var, cnt, alt],
        returns_scalar=True,
    )


def ttest_ind_from_stats(
    var1: str | pl.Expr,
    mean: float,
    var: float,
    cnt: int,
    alternative: Alternative = "two-sided",
    equal_var: bool = False,
) -> pl.Expr:
    """
    Performs 2 sample student's t test or Welch's t test, using only scalar statistics from other.
    This is more suitable for t-tests between rolling data and some other fixed data, from which you
    can compute the mean, var, and count only once.

    If (NaN, NaN) is returned, then it is possible that one of the following numeric
    problems occurred:

    1. There is no valid value in the inputs, or the mean is inf.
    2. Input variable has length 0 after removing non-finite values.

    Parameters
    ----------
    var1
        The variable 1
    mean
        The mean of var2
    var
        The var of var2
    cnt
        The count of var2, used only in welch's t test
    alternative : {"two-sided", "less", "greater"}
        Alternative of the hypothesis test
    equal_var
        If true, perform standard student t 2 sample test. Otherwise, perform Welch's
        t test.
    """
    y = to_expr(var1)
    if equal_var:
        m1 = y.mean()
        m2 = pl.lit(mean, pl.Float64)
        v1 = y.var()
        v2 = pl.lit(var, pl.Float64)
        cnt = y.count().cast(pl.UInt64)
        return pl_plugin(
            symbol="pl_ttest_2samp",
            args=[m1, m2, v1, v2, cnt, pl.lit(alternative, dtype=pl.String)],
            returns_scalar=True,
        )
    else:
        s1 = y.filter(y.is_finite())
        m1 = s1.mean()
        m2 = pl.lit(mean, pl.Float64)
        v1 = s1.var()
        v2 = pl.lit(var, pl.Float64)
        n1 = s1.len().cast(pl.UInt64)
        n2 = pl.lit(cnt, pl.UInt64)
        return pl_plugin(
            symbol="pl_welch_t",
            args=[m1, m2, v1, v2, n1, n2, pl.lit(alternative, dtype=pl.String)],
            returns_scalar=True,
        )


def ks_2samp(
    var1: str | pl.Expr,
    var2: str | pl.Expr,
    alpha: float = 0.05,
    is_binary: bool = False,
) -> pl.Expr:
    """
    Computes two-sided KS statistics between var1 and var2. This will
    sanitize data (only non-null finite values are used) before doing the computation. If
    is_binary is true, it will compare the statistics by comparing var2(var1=0) and var2(var1=1).

    Note, this returns a stastics and a threshold value. The threshold is not the p-value, but
    rather it is used in the following way: if the statistic is > the threshold value, then the null
    hypothesis should be rejected. This is suitable only for large sameple sizes. See more details
    in the reference.

    If either var1 or var2 has less than 30 values, a ks stats of 0 with threshold NaN will be returned.

    Parameters
    ----------
    var1
        Variable 1
    var2
        Variable 2
    alpha
        The confidence level used to estimate p-value
    is_binary
        If true, instead of running ks(var1, var2), it runs ks(var2(var1=0), var2(var1=1))

    Reference
    ---------
    https://en.wikipedia.org/wiki/Kolmogorov%E2%80%93Smirnov_test#Two-sample_Kolmogorov%E2%80%93Smirnov_test
    """
    y1, y2 = to_expr(var1), to_expr(var2)
    if is_binary:
        z1 = y2.filter(y1 == 1)
        z2 = y2.filter(y1 == 0)
        z1 = z1.filter(z1.is_finite()).sort()
        z2 = z2.filter(z2.is_finite()).sort()
    else:
        z1 = y1.filter(y1.is_finite()).sort()
        z2 = y2.filter(y2.is_finite()).sort()

    return pl_plugin(
        symbol="pl_ks_2samp",
        args=[z1.cast(pl.Float64), z2.cast(pl.Float64), pl.lit(alpha, pl.Float64)],
        returns_scalar=True,
    )


def f_test(*variables: str | pl.Expr, group: str | pl.Expr) -> pl.Expr:
    """
    Performs the ANOVA F-test.

    Parameters
    ----------
    variables
        The columns (variables) to run ANOVA F-test on
    group
        The "target" column used to group the variables
    """
    vars_ = [to_expr(group)]
    vars_.extend(to_expr(x) for x in variables)
    if len(vars_) <= 1:
        raise ValueError("No input feature column to run F-test on.")
    elif len(vars_) == 2:
        return pl_plugin(symbol="pl_f_test", args=vars_, returns_scalar=True)
    else:
        return pl_plugin(symbol="pl_f_test", args=vars_, changes_length=True)


def chi2(var1: str | pl.Expr, var2: str | pl.Expr, return_full: bool = False) -> pl.Expr:
    """
    Computes the Chi Squared statistic and p value between two categorical values.

    Note that it is up to the user to make sure that the two columns contain categorical
    values. This method is equivalent to SciPy's chi2_contingency, except that it also
    computes the contingency table internally for the user.

    Parameters
    ----------
    var1
        Either the name of the column or a Polars expression
    var2
        Either the name of the column or a Polars expression
    return_full
        If true, dof and expected frequency will also be returned. The returned "struct"
        will not be a scalar anymore, but has length = length of expected frequencies.
    """
    if return_full:
        return pl_plugin(
            symbol="pl_chi2_full", args=[to_expr(var1), to_expr(var2)], changes_length=True
        )
    else:
        return pl_plugin(
            symbol="pl_chi2",
            args=[to_expr(var1), to_expr(var2)],
            returns_scalar=True,
        )


def mann_whitney_u(
    var1: str | pl.Expr,
    var2: str | pl.Expr,
    alternative: Alternative = "two-sided",
) -> pl.Expr:
    """
    Computes the Mann-Whitney U statistic and the p-value. Note: this function will sanitize data (drop
    all non-finite values) before computing the statistic. This implementation follows method 2 in reference.
    This always applies tie correction, which may slow down computation by a little.

    WIP. PVALUE NOT DONE YET.

    Parameters
    ----------
    var1 : pl.Expr
        Either the name of the column or a Polars expression
    var2 : pl.Expr
        Either the name of the column or a Polars expression
    alternative: str
        The alternative for the test. `two-sided`, `greater` or `less`

    Reference
    ---------
    https://en.wikipedia.org/wiki/Mann%E2%80%93Whitney_U_test
    """
    x = to_expr(var1)
    y = to_expr(var2)
    xx = x.filter(x.is_finite())
    yy = y.filter(y.is_finite())
    n1 = xx.len().cast(pl.Float64)
    n2 = yy.len().cast(pl.Float64)

    ranks = (xx.append(yy)).rank()

    u1 = ranks.slice(0, length=xx.len()).sum() - (n1 * (n1 + 1)) / 2
    u2 = (n1 * n2) - u1

    mean = (n1 * n2) / 2
    return pl_plugin(
        symbol="pl_mann_whitney_u",
        args=[u1, u2, mean, ranks.sort(), pl.lit(alternative, dtype=pl.String)],
    )


def winsorize(
    x: str | pl.Expr,
    q_low: float = 0.05,
    q_high: float = 0.95,
    method: QuantileMethod = "nearest",
) -> pl.Expr:
    """
    Winsorize the data by clipping by percentiles at the lower and upper ends.

    Parameters
    ----------
    x
        Either the name of the column or a Polars expression
    q_low
        The lower percentile value to clip the data. E.g everything < x.quantile(lower)
        will be mapped to x.quantile(lower)
    q_high
        The upper percentile value to clip the data. E.g everything > x.quantile(upper)
        will be mapped to x.quantile(upper)
    method
        Method for quantile estimate. One of "nearest", "higher", "lower", "midpoint", "linear".
    """
    if q_low <= 0.0 or q_low >= 1.0 or q_high <= 0.0 or q_high >= 1.0 or q_high <= q_low:
        raise ValueError("Lower and upper must be with in (0, 1) and upper should be > lower")

    xx = to_expr(x)
    return xx.clip(
        xx.quantile(q_low, interpolation=method), xx.quantile(q_high, interpolation=method)
    )


def perturb(
    x: str | pl.Expr, epsilon: float = 1e-5, positive: bool = False, seed: int | None = None
) -> pl.Expr:
    """
    Perturb the var by a small amount. This only applies to float columns.

    Parameters
    ----------
    x
        Either the name of the column or a Polars expression
    epsilon
        The small amount to perturb.
    positive
        If true, randomly add a small amount in [0, epsilon). If false, it will use the range
        [-epsilon/2, epsilon/2)
    seed
        A random seed
    """
    if math.isinf(epsilon) or math.isnan(epsilon):
        raise ValueError("Input `epsilon should be a valid finite value.`")

    ep = abs(epsilon)
    if positive:
        lo = pl.lit(0.0, dtype=pl.Float64)
        hi = pl.lit(ep, dtype=pl.Float64)
    else:
        half = ep / 2
        lo = pl.lit(-half, dtype=pl.Float64)
        hi = pl.lit(half, dtype=pl.Float64)

    return pl_plugin(
        symbol="pl_perturb",
        args=[to_expr(x), lo, hi, pl.lit(seed, dtype=pl.UInt64)],
        is_elementwise=True,
    )


def jitter(x: str | pl.Expr, std: float | pl.Expr = 1.0, seed: int | None = None) -> pl.Expr:
    """
    Adds a Gaussian noise of N(0, std) to the column.

    Parameters
    ----------
    x
        Either the name of the column or a Polars expression
    std
        The std of the Gaussian noise.
    seed
        A random seed
    """
    if isinstance(std, float):
        if std < 0:
            raise ValueError("Standard deviation must be positive.")
        elif std == 0:
            return to_expr(x)

        s = pl.lit(std, dtype=pl.Float64)
    else:
        s = std.cast(pl.Float64)

    return pl_plugin(
        symbol="pl_jitter",
        args=[to_expr(x), s, pl.lit(seed, dtype=pl.UInt64)],
        is_elementwise=True,
    )


def add_noise(x: str | pl.Expr, noise_type: Noise = "gaussian", **kwargs) -> pl.Expr:
    """
    Adds some noise to the column.

    Parameters
    ----------
    x
        Either the name of the column or a Polars expression
    noise_type
        Either "gaussian" or "uniform"
    kwargs
        If noise_type = "gaussian", this accepts kwargs to "jitter" and if "uniform", this
        accepts kwargs to "perturb". You may set a seed via the kwargs.
    """
    if noise_type == "gaussian":
        return jitter(x, **kwargs)
    elif noise_type == "uniform":
        return perturb(x, **kwargs)
    else:
        raise ValueError(f"The noise_type {noise_type} is not currently supported.")


def normal_test(var: str | pl.Expr) -> pl.Expr:
    """
    Perform a normality test which is based on D'Agostino and Pearson's test
    that combines skew and kurtosis to produce an omnibus test of normality.
    Null values, NaN and inf are dropped when running this computation.

    Parameters
    ----------
    var
        Either the name of the column or a Polars expression

    References
    ----------
    D'Agostino, R. B. (1971), "An omnibus test of normality for
        moderate and large sample size", Biometrika, 58, 341-348
    D'Agostino, R. and Pearson, E. S. (1973), "Tests for departure from
        normality", Biometrika, 60, 613-622
    """
    y = to_expr(var)
    valid: pl.Expr = y.filter(y.is_finite())
    skew = valid.skew()
    # Pearson Kurtosis, see here: https://en.wikipedia.org/wiki/D%27Agostino%27s_K-squared_test
    kur = valid.kurtosis(fisher=False)
    return pl_plugin(
        symbol="pl_normal_test",
        args=[skew, kur, valid.count().cast(pl.UInt32)],
        returns_scalar=True,
    )


def random(
    lower: pl.Expr | float = 0.0,
    upper: pl.Expr | float = 1.0,
    seed: int | None = None,
    length: int | pl.Expr = pl.len(),
) -> pl.Expr:
    """
    Generate random numbers in [lower, upper)

    Parameters
    ----------
    lower
        The lower bound
    upper
        The upper bound, exclusive
    seed
        The random seed. None means no seed.
    length
        Custom length. Note length needs to match with other columns in the context.
    """
    lo = pl.lit(lower, pl.Float64) if isinstance(lower, float) else lower
    up = pl.lit(upper, pl.Float64) if isinstance(upper, float) else upper
    len_ = pl.lit(length, pl.UInt32) if isinstance(length, int) else length
    return pl_plugin(
        symbol="pl_random",
        args=[len_, lo, up, pl.lit(seed, pl.UInt64)],
        is_elementwise=True,
    )


def random_null(x: str | pl.Expr, pct: float, seed: int | None = None) -> pl.Expr:
    """
    Creates random null values in the columns. If var contains nulls originally, they
    will stay null.

    Parameters
    ----------
    x
        Either the name of the column or a Polars expression
    pct
        Percentage of nulls to randomly generate. This percentage is based on the
        length of the column, so may not be the actual percentage of nulls depending
        on how many values are originally null.
    seed
        A seed to fix the random numbers. If none, use the system's entropy.
    """
    if pct <= 0.0 or pct >= 1.0:
        raise ValueError("Input `pct` must be > 0 and < 1")

    return pl.when(random(0.0, 1.0, seed=seed) < pct).then(None).otherwise(to_expr(x))


def random_int(
    lower: int | pl.Expr,
    upper: int | pl.Expr,
    seed: int | None = None,
    length: int | pl.Expr = pl.len(),
) -> pl.Expr:
    """
    Generates random integer between lower and upper.

    Parameters
    ----------
    lower
        The lower bound, inclusive
    upper
        The upper bound, exclusive
    seed
        The random seed. None means no seed.
    length
        Custom length. Note length needs to match with other columns in the context.
    """
    if lower == upper:
        raise ValueError("Input `lower` must be smaller than `higher`")

    lo = pl.lit(lower, pl.Int32) if isinstance(lower, int) else lower.cast(pl.Int32)
    hi = pl.lit(upper, pl.Int32) if isinstance(upper, int) else upper.cast(pl.Int32)
    len_ = pl.lit(length, pl.UInt32) if isinstance(length, int) else length
    return pl_plugin(
        symbol="pl_rand_int",
        args=[
            len_,
            lo,
            hi,
            pl.lit(seed, pl.UInt64),
        ],
        is_elementwise=True,
    )


def random_str(
    min_size: int,
    max_size: int,
    seed: int | None = None,
    length: int | pl.Expr = pl.len(),
) -> pl.Expr:
    """
    Generates random strings of length between min_size and max_size. The characters are
    uniformly distributed over ASCII letters and numbers: a-z, A-Z and 0-9.

    Parameters
    ----------
    min_size
        The min size of the string, inclusive
    max_size
        The max size of the string, inclusive
    seed
        The random seed. None means no seed.
    length
        Custom length. Note length needs to match with other columns in the context.
    """
    mi, ma = min_size, max_size
    if min_size > max_size:
        mi, ma = max_size, min_size

    return pl_plugin(
        symbol="pl_rand_str",
        args=[
            pl.lit(length, pl.UInt32) if isinstance(length, int) else length,
            pl.lit(mi, pl.UInt32),
            pl.lit(ma, pl.UInt32),
            pl.lit(seed, pl.UInt64),
        ],
        is_elementwise=True,
    )


def random_binomial(
    n: int, p: int, seed: int | None = None, length: int | pl.Expr = pl.len()
) -> pl.Expr:
    """
    Generates random integer following a binomial distribution.

    Parameters
    ----------
    n
        The n in a binomial distribution
    p
        The p in a binomial distribution
    seed
        The random seed. None means no seed.
    length
        Custom length. Note length needs to match with other columns in the context.
    """
    if n < 1:
        raise ValueError("Input `n` must be > 1.")

    return pl_plugin(
        symbol="pl_rand_binomial",
        args=[
            pl.lit(length, pl.UInt32) if isinstance(length, int) else length,
            pl.lit(n, pl.Int32),
            pl.lit(p, pl.Float64),
            pl.lit(seed, pl.UInt64),
        ],
        is_elementwise=True,
    )


def random_exp(
    lambda_: float, seed: int | None = None, length: int | pl.Expr = pl.len()
) -> pl.Expr:
    """
    Generates random numbers following an exponential distribution.

    Parameters
    ----------
    lambda_
        The lambda in an exponential distribution
    seed
        The random seed. None means no seed.
    length
        Custom length. Note length needs to match with other columns in the context.
    """
    return pl_plugin(
        symbol="pl_rand_exp",
        args=[
            pl.lit(length, pl.UInt32) if isinstance(length, int) else length,
            pl.lit(lambda_, pl.Float64),
            pl.lit(seed, pl.UInt64),
        ],
        is_elementwise=True,
    )


def random_normal(
    mean: pl.Expr | float,
    std: pl.Expr | float,
    seed: int | None = None,
    length: int | pl.Expr = pl.len(),
) -> pl.Expr:
    """
    Generates random number following a normal distribution.

    Parameters
    ----------
    mean
        The mean in a normal distribution
    std
        The std in a normal distribution
    seed
        The random seed. None means no seed.
    length
        Custom length. Note length needs to match with other columns in the context.
    """
    return pl_plugin(
        symbol="pl_rand_normal",
        args=[
            pl.lit(length, pl.UInt32) if isinstance(length, int) else length,
            pl.lit(mean, pl.Float64) if isinstance(mean, float) else mean,
            pl.lit(std, pl.Float64) if isinstance(std, float) else std,
            pl.lit(seed, pl.UInt64),
        ],
        is_elementwise=True,
    )


def hmean(var: str | pl.Expr) -> pl.Expr:
    """
    Computes the harmonic mean.

    Parameters
    ----------
    var
        The variable
    """
    x = to_expr(var)
    return x.count() / (1.0 / x).sum()


def weighted_hmean(
    var: str | pl.Expr, weights: str | pl.Expr, is_normalized: bool = False
) -> pl.Expr:
    """
    Computes the weighted harmonic mean of the variable.

    Parameters
    ----------
    var
        The variable
    weights
        An expr representing weights. Must be of same length as var.
    is_normalized
        If true, the weights are assumed to sum to 1. If false, will divide by sum of the weights
    """
    w = to_expr(weights)
    x = to_expr(var)
    dot = x.dot(pl.lit(1.0, dtype=pl.Float32) / x)
    if is_normalized:
        return 1.0 / dot
    else:
        return 1.0 / (dot / w.sum())


def gmean(var: str | pl.Expr) -> pl.Expr:
    """
    Computes the geometric mean of the variable.

    Parameters
    ----------
    var
        The variable
    """
    return to_expr(var).ln().mean().exp()


def weighted_gmean(
    var: str | pl.Expr, weights: str | pl.Expr, is_normalized: bool = False
) -> pl.Expr:
    """
    Computes the weighted geometric mean of the variable.

    Parameters
    ----------
    var
        The variable
    weights
        An expr representing weights. Must be of same length as var.
    is_normalized
        If true, the weights are assumed to sum to 1. If false, will divide by sum of the weights
    """
    x, w = to_expr(var), to_expr(weights)
    if is_normalized:
        return (x.ln().dot(w)).exp()
    else:
        return (x.ln().dot(w) / (w.sum())).exp()


def weighted_mean(
    var: str | pl.Expr, weights: str | pl.Expr, is_normalized: bool = False
) -> pl.Expr:
    """
    Computes the weighted mean, where weights is an expr represeting
    a weight column. The weights column must have the same length as var.

    All weights are assumed to be > 0. This will not check if weights are valid.

    Parameters
    ----------
    var
        The variable
    weights
        An expr representing weights. Must be of same length as var.
    is_normalized
        If true, the weights are assumed to sum to 1. If false, will divide by sum of the weights
    """
    x, w = to_expr(var), to_expr(weights)
    out = x.dot(w)
    if is_normalized:
        return out
    return out / w.sum()


def weighted_var(var: str | pl.Expr, weights: str | pl.Expr, freq_weights: bool = False) -> pl.Expr:
    """
    Computes the weighted variance. The weights column must have the same length as var.

    All weights are assumed to be > 0. This will not check if weights are valid.

    Parameters
    ----------
    var
        The variable
    weights
        An expr representing weights. Must be of same length as var.
    freq_weights
        Whether to follow the formula for frequency weights or other types of weights. See reference
        for detail. If true, this assumes frequency weights are NOT normalized. If false, the
        weighted sample variance is biased. See reference for more info.

    Reference
    ---------
    https://en.wikipedia.org/wiki/Weighted_arithmetic_mean#Weighted_sample_variance
    """
    x, w = to_expr(var), to_expr(weights)
    wm = weighted_mean(x, w, False)
    summand = w.dot((x - wm).pow(2))
    if freq_weights:
        return summand / (w.sum() - 1)
    return summand / w.sum()


def weighted_cov(x: str | pl.Expr, y: str | pl.Expr, weights: pl.Expr | float) -> pl.Expr:
    """
    Computes the weighted covariance between x and y. The weights column must have the same
    length as both x an y.

    All weights are assumed to be > 0. This will not check if weights are valid.

    Parameters
    ----------
    x
        The first variable
    y
        The second variable
    weights
        An expr representing weights. Must be of same length as var.

    Reference
    ---------
    https://en.wikipedia.org/wiki/Pearson_correlation_coefficient#Weighted_correlation_coefficient
    """
    xx, yy, w = to_expr(x), to_expr(y), to_expr(weights)
    wx, wy = weighted_mean(xx, w, False), weighted_mean(yy, w, False)
    return w.dot((xx - wx) * (yy - wy)) / w.sum()


def weighted_corr(x: str | pl.Expr, y: str | pl.Expr, weights: str | pl.Expr) -> pl.Expr:
    """
    Computes the weighted correlation between x and y. The weights column must have the same
    length as both x an y.

    All weights are assumed to be > 0. This will not check if weights are valid.

    Parameters
    ----------
    x
        The first variable
    y
        The second variable
    weights
        An expr representing weights. Must be of same length as var.

    Reference
    ---------
    https://en.wikipedia.org/wiki/Pearson_correlation_coefficient#Weighted_correlation_coefficient
    """
    xx, yy = to_expr(x), to_expr(y)
    w = to_expr(weights)
    numerator = w.dot((xx - weighted_mean(xx, w, False)) * (yy - weighted_mean(yy, w, False)))
    sxx = w.dot((xx - weighted_mean(xx, w, False)).pow(2))
    syy = w.dot((yy - weighted_mean(yy, w, False)).pow(2))
    return numerator / (sxx * syy).sqrt()


def cosine_sim(x: str | pl.Expr, y: str | pl.Expr) -> pl.Expr:
    """
    Column-and-column cosine similarity

    Parameters
    ----------
    x
        The first variable
    y
        The second variable
    """
    xx, yy = to_expr(x), to_expr(y)
    x2 = xx.dot(xx).sqrt()
    y2 = yy.dot(yy).sqrt()
    return xx.dot(yy) / (x2 * y2).sqrt()


def weighted_cosine_sim(x: str | pl.Expr, y: str | pl.Expr, weights: str | pl.Expr) -> pl.Expr:
    """
    Computes the weighted cosine similarity between x and y (column-wise). The weights column
    must have the same length as both x an y.

    All weights are assumed to be > 0. This will not check if weights are valid.

    Parameters
    ----------
    x
        The first variable
    y
        The second variable
    weights
        An expr representing weights. Must be of same length as var.

    Reference
    ---------
    https://en.wikipedia.org/wiki/Pearson_correlation_coefficient#Weighted_correlation_coefficient
    """
    xx, yy = to_expr(x), to_expr(y)
    w = to_expr(weights)
    wx2 = xx.pow(2).dot(w)
    wy2 = yy.pow(2).dot(w)
    return (w * xx).dot(yy) / (wx2 * wy2).sqrt()


def kendall_tau(x: str | pl.Expr, y: str | pl.Expr) -> pl.Expr:
    """
    Computes Kendall's Tau (b) correlation between x and y. This automatically drops rows with null.

    Note: this will map NaN to null and drop all rows with null. Inf will be kept and cosidered as
    the largest value and multiple Infs will be equal. -Inf will be the smallest if it exists in the
    data. A value of NaN will be returned if the data has < 2 rows after nulls are dropped.

    Parameters
    ----------
    x
        The first variable
    y
        The second variable
    """
    xx, yy = to_expr(x).fill_nan(None), to_expr(y).fill_nan(None)
    return pl_plugin(
        symbol="pl_kendall_tau",
        args=[xx.rank(method="min"), yy.rank(method="min")],
        returns_scalar=True,
    )


def bicor(x: str | pl.Expr, y: str | pl.Expr, c: float = 9.0) -> pl.Expr:
    """
    Computes the Biweight Midcorrelation between x and y. This is commonly referred to as bicor.

    Performance hint: this expression benefits from .lazy() a lot.

    Parameters
    ----------
    x
        The first variable
    y
        The second variable
    c
        Biweight tuning constant which is typically 9

    Reference
    ---------
    https://en.wikipedia.org/wiki/Biweight_midcorrelation
    """
    a, b = to_expr(x), to_expr(y)
    med_a = a.median()
    med_b = b.median()

    diff_a = a - med_a
    diff_b = b - med_b

    ua = diff_a / (c * diff_a.abs().median())
    ub = diff_b / (c * diff_b.abs().median())

    w_a = (1 - ua.pow(2)).pow(2) * ((1 - ua.abs()) > 0).cast(pl.Float64)
    w_b = (1 - ub.pow(2)).pow(2) * ((1 - ub.abs()) > 0).cast(pl.Float64)

    aa = diff_a * w_a
    bb = diff_b * w_b

    return aa.dot(bb) / (aa.dot(aa) * (bb.dot(bb))).sqrt()


def xi_corr(
    x: str | pl.Expr, y: str | pl.Expr, seed: int | None = None, return_p: bool = False
) -> pl.Expr:
    """
    Computes the ξ(xi) correlation developed by SOURAV CHATTERJEE in the paper in the reference.
    This will return both the correlation (the statistic) and the p-value. Note that if sample size
    is smaller than 30, p-value will always be NaN. The ξ correlation is not symmetric, as it only
    tries to explain whether y is a function of x.

    Parameters
    ----------
    x
        The first variable
    y
        The second variable
    seed
        Whether to have a seed when we break ties at random
    return_p
        Whether to return a two-sided p value for the statistic

    Reference
    ---------
    https://arxiv.org/pdf/1909.10140.pdf
    """
    xx, yy = to_expr(x), to_expr(y)
    args = [
        xx.rank(method="random", seed=seed),
        yy.rank(method="max").cast(pl.Float64),
        (-yy).rank(method="max").cast(pl.Float64),
    ]
    if return_p:
        return pl_plugin(
            symbol="pl_xi_corr_w_p",
            args=args,
            returns_scalar=True,
        )
    else:
        return pl_plugin(
            symbol="pl_xi_corr",
            args=args,
            returns_scalar=True,
        )


def corr(x: str | pl.Expr, y: str | pl.Expr, method: CorrMethod = "pearson") -> pl.Expr:
    """
    A convenience function for calling different types of correlations. Pearson and Spearman correlation
    runs on Polar's native expression, while Kendall and Xi correlation runs on code in this package.

    Paramters
    ---------
    x
        The first variable
    y
        The second variable
    method
        One of ["pearson", "spearman", "xi", "kendall", "bicor"]
    """
    if method in ["pearson", "spearman"]:
        return pl.corr(x, y, method=method)
    elif method == "xi":
        return xi_corr(x, y)
    elif method == "kendall":
        return kendall_tau(x, y)
    elif method == "bicor":
        return bicor(x, y)
    else:
        raise ValueError(f"Unknown correlation method: {method}.")



from __future__ import annotations
from typing import Callable
import polars as pl


def E(
    cols: str | pl.Expr | list[str] | list[pl.Expr],
    mappings: str | list[str],
    *,
    separator: str = "_",
    len_alias: str = "__len__",
    customizer: dict[str, Callable[[pl.Expr], pl.Expr]] | None = None,
) -> list[pl.Expr]:
    """
    Automatically expands Polars expressions so that the syntax can be more concise
    in most cases.

    Parameters
    ----------
    cols
        Either the column name list of column names. Due to limitations in Polars,
        this does not support expression inputs.
    mappings
        String names for any polars expression methods. E.g. ['max'] will map to `pl.col(..).max()`.
        There are some special strings that would represent some more complicated composite expressions,
        and here is a the list of such special mappings as of now:
        'null_rate' -> pl.col('').null_count() / pl.len()
    separator
        The separator between the name of the mapping and the original column name.
    len_alias
        Column alias to `len`.
    customizer
        An additional mapping to add your own custom special mappings. The keys are the name
        of additional mappings, the values are functions that maps pl.Expr to another pl.Expr.

    Examples
    --------
    >>> import polars_ds as pds
    >>> df = pl.DataFrame({
    >>>     "group": ['A', 'A', 'B', 'B', 'A']
    >>>     , "a": [1, 2, 3, 4, 5]
    >>>     , "b": [4, 1, 99, 12, 33]
    >>> })
    >>> df.group_by("group").agg(
    >>>     *pds.E(['a', 'b'], ["min", "max", "n_unique", "len"])
    >>> )
    shape: (2, 8)
    ┌───────┬───────┬───────┬───────┬───────┬────────────┬────────────┬─────────┐
    │ group ┆ a_min ┆ b_min ┆ a_max ┆ b_max ┆ a_n_unique ┆ b_n_unique ┆ __len__ │
    │ ---   ┆ ---   ┆ ---   ┆ ---   ┆ ---   ┆ ---        ┆ ---        ┆ ---     │
    │ str   ┆ i64   ┆ i64   ┆ i64   ┆ i64   ┆ u32        ┆ u32        ┆ u32     │
    ╞═══════╪═══════╪═══════╪═══════╪═══════╪════════════╪════════════╪═════════╡
    │ A     ┆ 1     ┆ 1     ┆ 5     ┆ 33    ┆ 3          ┆ 3          ┆ 3       │
    │ B     ┆ 3     ┆ 12    ┆ 4     ┆ 99    ┆ 2          ┆ 2          ┆ 2       │
    └───────┴───────┴───────┴───────┴───────┴────────────┴────────────┴─────────┘

    >>> customizer = {
    >>>     "quantile_90": lambda x: x.quantile(0.9).name.suffix("_q90")
    >>> }
    >>> df.group_by("group").agg(
    >>>     *pds.E(['a', 'b'], ["min", "max", "null_rate", "quantile_90"], customizer=customizer)
    >>> )
    shape: (2, 9)
    ┌───────┬───────┬───────┬───────┬───┬─────────────┬─────────────┬───────┬───────┐
    │ group ┆ a_min ┆ b_min ┆ a_max ┆ … ┆ a_null_rate ┆ b_null_rate ┆ a_q90 ┆ b_q90 │
    │ ---   ┆ ---   ┆ ---   ┆ ---   ┆   ┆ ---         ┆ ---         ┆ ---   ┆ ---   │
    │ str   ┆ i64   ┆ i64   ┆ i64   ┆   ┆ f64         ┆ f64         ┆ f64   ┆ f64   │
    ╞═══════╪═══════╪═══════╪═══════╪═══╪═════════════╪═════════════╪═══════╪═══════╡
    │ A     ┆ 1     ┆ 1     ┆ 5     ┆ … ┆ 0.0         ┆ 0.0         ┆ 5.0   ┆ 33.0  │
    │ B     ┆ 3     ┆ 12    ┆ 4     ┆ … ┆ 0.0         ┆ 0.0         ┆ 4.0   ┆ 99.0  │
    └───────┴───────┴───────┴───────┴───┴─────────────┴─────────────┴───────┴───────┘
    """

    if isinstance(cols, str):
        columns = [cols]
    elif isinstance(cols, list):
        if any(not isinstance(c, str) for c in cols):
            raise TypeError(
                "Input `cols` must either be a single str/pl.Expr or a list of str/pl.expr."
            )
        columns = list(cols)
    else:
        raise TypeError(
            "Input `cols` must either be a single str/pl.Expr or a list of str/pl.expr."
        )

    in_expr = pl.col(columns)

    if isinstance(mappings, str):
        mappings_ = [mappings]
    else:
        mappings_ = list(mappings)

    SPECIAL_MAPPINGS = {
        "len": pl.len().alias(len_alias),
        "count": pl.len().alias(len_alias),
        "null_rate": (in_expr.null_count() / pl.len()).name.suffix(f"{separator}null_rate"),
    }

    if customizer is not None and isinstance(customizer, dict):
        for k in customizer.keys():
            SPECIAL_MAPPINGS[k] = customizer[k](in_expr)

    bad = [m for m in mappings_ if not (hasattr(in_expr, m) or m in SPECIAL_MAPPINGS)]
    if len(bad) > 0:
        raise ValueError(
            f"Polars expressions does not have `{bad}` method(s) and they do not belong "
            "in the special mappings."
        )

    return [
        SPECIAL_MAPPINGS[m]
        if m in SPECIAL_MAPPINGS
        else getattr(in_expr, m)().name.suffix(f"{separator}{m}")
        for m in mappings_
    ]
